import datetime as dt
import logging
from typing import List, Optional

import petname

from cg.constants import DataDelivery, Pipeline, FlowCellStatus
from cg.store.api.base import BaseHandler

from cg.constants import Priority
from cg.store.models import (
    Flowcell,
    Organism,
    Customer,
    Sample,
    Pool,
    Delivery,
    ApplicationVersion,
    Panel,
    Analysis,
    Family,
    FamilySample,
    Bed,
    BedVersion,
    Application,
    User,
    Collaboration,
)

LOG = logging.getLogger(__name__)


class AddHandler(BaseHandler):
    """Methods related to adding new data to the store."""

    def generate_unique_petname(self) -> str:
        while True:
            random_id = petname.Generate(3, separator="")
            if not self.get_sample_by_internal_id(internal_id=random_id):
                return random_id

    def add_customer(
        self,
        internal_id: str,
        name: str,
        invoice_address: str,
        invoice_reference: str,
        scout_access: bool = False,
        *args,
        **kwargs,
    ) -> Customer:
        """Build a new customer record."""

        return self.Customer(
            internal_id=internal_id,
            name=name,
            scout_access=scout_access,
            invoice_address=invoice_address,
            invoice_reference=invoice_reference,
            **kwargs,
        )

    def add_collaboration(self, internal_id: str, name: str, **kwargs) -> Collaboration:
        """Build a new customer group record."""

        return self.Collaboration(internal_id=internal_id, name=name, **kwargs)

    def add_user(self, customer: Customer, email: str, name: str, is_admin: bool = False) -> User:
        """Build a new user record."""

        new_user = self.User(name=name, email=email, is_admin=is_admin)
        new_user.customers.append(customer)
        return new_user

    def add_application(
        self,
        tag: str,
        prep_category: str,
        description: str,
        percent_kth: int,
        percent_reads_guaranteed: int,
        is_accredited: bool = False,
        **kwargs,
    ) -> Application:
        """Build a new application  record."""

        return self.Application(
            tag=tag,
            prep_category=prep_category,
            description=description,
            is_accredited=is_accredited,
            percent_kth=percent_kth,
            percent_reads_guaranteed=percent_reads_guaranteed,
            **kwargs,
        )

    def add_application_version(
        self,
        application: Application,
        version: int,
        valid_from: dt.datetime,
        prices: dict,
        **kwargs,
    ) -> ApplicationVersion:
        """Build a new application version record."""

        new_record = self.ApplicationVersion(version=version, valid_from=valid_from, **kwargs)
        for price_key in [
            Priority.standard.name,
            Priority.priority.name,
            Priority.express.name,
            Priority.research.name,
        ]:
            setattr(new_record, f"price_{price_key}", prices[price_key])
        new_record.application = application
        return new_record

    def add_bed(self, name: str) -> Bed:
        """Build a new bed record."""
        return self.Bed(name=name)

    def add_bed_version(self, bed: Bed, version: int, filename: str, shortname: str) -> BedVersion:
        """Build a new bed version record."""
        bed_version: BedVersion = self.BedVersion(
            version=version, filename=filename, shortname=shortname
        )
        bed_version.bed = bed
        return bed_version

    def add_sample(
        self,
        name: str,
        sex: str,
        comment: str = None,
        control: str = None,
        downsampled_to: int = None,
        internal_id: str = None,
        order: str = None,
        ordered: dt.datetime = None,
        priority: Priority = None,
        received: dt.datetime = None,
        original_ticket: str = None,
        tumour: bool = False,
        **kwargs,
    ) -> Sample:
        """Build a new Sample record."""

        internal_id = internal_id or self.generate_unique_petname()
        priority = priority or (Priority.research if downsampled_to else Priority.standard)
        return self.Sample(
            comment=comment,
            control=control,
            downsampled_to=downsampled_to,
            internal_id=internal_id,
            is_tumour=tumour,
            name=name,
            order=order,
            ordered_at=ordered or dt.datetime.now(),
            original_ticket=original_ticket,
            priority=priority,
            received_at=received,
            sex=sex,
            **kwargs,
        )

    def add_case(
        self,
        data_analysis: Pipeline,
        data_delivery: DataDelivery,
        name: str,
        ticket: str,
        panels: Optional[List[str]] = None,
        cohorts: Optional[List[str]] = None,
        priority: Optional[Priority] = Priority.standard,
        synopsis: Optional[str] = None,
    ) -> Family:
        """Build a new Family record."""

        # generate a unique case id
        while True:
            internal_id = petname.Generate(2, separator="")
            if self.get_case_by_internal_id(internal_id) is None:
                break
            else:
                LOG.debug(f"{internal_id} already used - trying another id")

        return self.Family(
            cohorts=cohorts,
            data_analysis=str(data_analysis),
            data_delivery=str(data_delivery),
            internal_id=internal_id,
            name=name,
            panels=panels,
            priority=priority,
            synopsis=synopsis,
            tickets=ticket,
        )

    def relate_sample(
        self,
        family: Family,
        sample: Sample,
        status: str,
        mother: Sample = None,
        father: Sample = None,
    ) -> FamilySample:
        """Relate a sample record to a family record."""

        new_record = self.FamilySample(status=status)
        new_record.family = family
        new_record.sample = sample
        new_record.mother = mother
        new_record.father = father
        return new_record

    def add_flow_cell(
        self,
        flow_cell_id: str,
        sequencer_name: str,
        sequencer_type: str,
        date: dt.datetime,
        flow_cell_status: Optional[str] = FlowCellStatus.ON_DISK,
    ) -> Flowcell:
        """Build a new Flowcell record."""
        return self.Flowcell(
            name=flow_cell_id,
            sequencer_name=sequencer_name,
            sequencer_type=sequencer_type,
            sequenced_at=date,
            status=flow_cell_status,
        )

    def add_analysis(
        self,
        pipeline: Pipeline,
        version: str = None,
        completed_at: dt.datetime = None,
        primary: bool = False,
        uploaded: dt.datetime = None,
        started_at: dt.datetime = None,
        **kwargs,
    ) -> Analysis:
        """Build a new Analysis record."""
        return self.Analysis(
            pipeline=str(pipeline),
            pipeline_version=version,
            completed_at=completed_at,
            is_primary=primary,
            uploaded_at=uploaded,
            started_at=started_at,
            **kwargs,
        )

    def add_panel(
        self,
        customer: Customer,
        name: str,
        abbrev: str,
        version: float,
        date: dt.datetime = None,
        genes: int = None,
    ) -> Panel:
        """Build a new panel record."""

        new_record = self.Panel(
            name=name, abbrev=abbrev, current_version=version, date=date, gene_count=genes
        )
        new_record.customer = customer
        return new_record

    def add_pool(
        self,
        customer: Customer,
        name: str,
        order: str,
        ordered: dt.datetime,
        application_version: ApplicationVersion,
        ticket: str = None,
        comment: str = None,
        received_at: dt.datetime = None,
        capture_kit: str = None,
        invoice_id: int = None,
        no_invoice: bool = None,
        delivered_at: dt.datetime = None,
    ) -> Pool:
        """Build a new Pool record."""

        new_record = self.Pool(
            name=name,
            ordered_at=ordered or dt.datetime.now(),
            order=order,
            ticket=ticket,
            received_at=received_at,
            comment=comment,
            capture_kit=capture_kit,
            delivered_at=delivered_at,
            invoice_id=invoice_id,
            no_invoice=no_invoice,
        )
        new_record.customer = customer
        new_record.application_version = application_version
        return new_record

    def add_delivery(
        self,
        destination: str,
        sample: Sample = None,
        pool: Pool = None,
        comment: str = None,
    ) -> Delivery:
        """Build a new Delivery record."""

        if not any([sample, pool]):
            raise ValueError("you have to provide a sample or a pool")
        new_record = self.Delivery(destination=destination, comment=comment)
        new_record.sample = sample
        new_record.pool = pool
        return new_record

    def add_invoice(
        self,
        customer: Customer,
        samples: List[Sample] = None,
        microbial_samples: List[Sample] = None,
        pools: List[Pool] = None,
        comment: str = None,
        discount: int = 0,
        record_type: str = None,
        invoiced_at: Optional[dt.datetime] = None,
    ):
        """Build a new Invoice record."""

        new_id = self.new_invoice_id()
        new_invoice = self.Invoice(
            comment=comment,
            discount=discount,
            id=new_id,
            record_type=record_type,
            invoiced_at=invoiced_at,
        )
        new_invoice.customer = customer
        for sample in samples or []:
            new_invoice.samples.append(sample)
        for microbial_sample in microbial_samples or []:
            new_invoice.samples.append(microbial_sample)
        for pool in pools or []:
            new_invoice.pools.append(pool)
        return new_invoice

    def add_organism(
        self,
        internal_id: str,
        name: str,
        reference_genome: str = None,
        verified: bool = False,
        **kwargs,
    ) -> Organism:
        """Build a new Organism record."""
        return self.Organism(
            internal_id=internal_id,
            name=name,
            reference_genome=reference_genome,
            verified=verified,
            **kwargs,
        )



"""Tests for GensAPI."""

from pathlib import Path
from typing import Dict

from cg.apps.gens import GensAPI
from cg.constants.gene_panel import GENOME_BUILD_37
from cg.utils.commands import Process


def test_instantiate(gens_config: Dict[str, Dict[str, str]]):
    """Test to instantiate GensAPI."""
    # GIVEN a Gens API config dictionary

    # WHEN instantiating a Gens API
    gens_api = GensAPI(gens_config)

    # THEN assert that the object was properly instantiated
    assert gens_api.process.config == gens_config["gens"]["config_path"]
    assert gens_api.process.binary == gens_config["gens"]["binary_path"]


def test_gens_api_load(
    case_id: str,
    gens_config: Dict[str, Dict[str, str]],
    gens_coverage_path: Path,
    gens_fracsnp_path: Path,
    mocker,
    sample_id: str,
):
    """Test load sample method."""
    # WHEN uploading a sample with the API, using a mocked Process.run_command method
    api: GensAPI = GensAPI(gens_config)
    mocked_run_command = mocker.patch.object(Process, "run_command")
    api.load(
        sample_id=sample_id,
        genome_build=GENOME_BUILD_37,
        baf_path=gens_fracsnp_path.as_posix(),
        coverage_path=gens_coverage_path.as_posix(),
        case_id=case_id,
    )

    # THEN run_command should be called with the list
    mocked_run_command.assert_called_once_with(
        dry_run=False,
        parameters=[
            "load",
            "sample",
            "--sample-id",
            sample_id,
            "--genome-build",
            GENOME_BUILD_37,
            "--baf",
            gens_fracsnp_path.as_posix(),
            "--coverage",
            gens_coverage_path.as_posix(),
            "--case-id",
            case_id,
        ],
    )



"""Fixtures for CLI tests."""
from pathlib import Path
from typing import Dict, Any

import pytest

from cg.apps.crunchy import CrunchyAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.meta.compress import CompressAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from click.testing import CliRunner

from tests.cli.compress.conftest import CaseInfo
from tests.store_helpers import StoreHelpers


@pytest.fixture(name="cli_runner")
def fixture_cli_runner() -> CliRunner:
    """Create a CliRunner"""
    return CliRunner()


@pytest.fixture(name="application_tag")
def fixture_application_tag() -> str:
    """Return a dummy tag"""
    return "dummy_tag"


@pytest.fixture(name="base_context")
def fixture_base_context(
    base_store: Store, housekeeper_api: HousekeeperAPI, cg_config_object: CGConfig
) -> CGConfig:
    """context to use in CLI."""
    cg_config_object.status_db_ = base_store
    cg_config_object.housekeeper_api_ = housekeeper_api
    return cg_config_object


@pytest.fixture(name="before_date")
def fixture_before_date() -> str:
    """Return a before date string"""
    return "1999-12-31"


@pytest.fixture(name="disk_store")
def fixture_disk_store(base_context: CGConfig) -> Store:
    """context to use in cli"""
    return base_context.status_db


class MockCompressAPI(CompressAPI):
    """Mock out necessary functions for running the compress CLI functions."""

    def __init__(self):
        """initialize mock"""
        super().__init__(hk_api=None, crunchy_api=None, demux_root="")
        self.ntasks = 12
        self.mem = 50
        self.fastq_compression_success = True
        self.spring_decompression_success = True
        self.dry_run = False

    def set_dry_run(self, dry_run: bool):
        """Update dry run."""
        self.dry_run = dry_run

    def compress_fastq(self, sample_id: str, dry_run: bool = False):
        """Return if compression was successful."""
        _ = sample_id, dry_run
        return self.fastq_compression_success

    def decompress_spring(self, sample_id: str, dry_run: bool = False):
        """Return if decompression was succesful."""
        _ = sample_id, dry_run
        return self.spring_decompression_success


@pytest.fixture(name="compress_api")
def fixture_compress_api():
    """Return a compress API context."""
    return MockCompressAPI()


@pytest.fixture(name="compress_context")
def fixture_base_compress_context(
    compress_api: CompressAPI, store: Store, cg_config_object: CGConfig
) -> CGConfig:
    """Return a compress context."""
    cg_config_object.meta_apis["compress_api"] = compress_api
    cg_config_object.status_db_ = store
    return cg_config_object


@pytest.fixture(name="compress_case_info")
def fixture_compress_case_info(
    case_id,
    family_name,
    timestamp,
    later_timestamp,
    wgs_application_tag,
):
    """Returns a object with information about a case."""
    return CaseInfo(
        case_id=case_id,
        family_name=family_name,
        timestamp=timestamp,
        later_timestamp=later_timestamp,
        application_tag=wgs_application_tag,
    )


@pytest.fixture(name="populated_compress_store")
def fixture_populated_compress_store(store, helpers, compress_case_info, analysis_family):
    """Return a store populated with a completed analysis."""
    # Make sure that there is a case where analysis is completed
    helpers.ensure_case_from_dict(
        store,
        case_info=analysis_family,
        app_tag=compress_case_info.application_tag,
        ordered_at=compress_case_info.timestamp,
        completed_at=compress_case_info.later_timestamp,
    )
    return store


@pytest.fixture(name="populated_compress_context")
def fixture_populated_compress_context(
    compress_api: CompressAPI, populated_compress_store: Store, cg_config_object: CGConfig
) -> CGConfig:
    """Return a compress context populated with a completed analysis."""
    # Make sure that there is a case where analysis is completed
    cg_config_object.meta_apis["compress_api"] = compress_api
    cg_config_object.status_db_ = populated_compress_store
    return cg_config_object


@pytest.fixture(name="real_crunchy_api")
def fixture_real_crunchy_api(crunchy_config: Dict[str, Dict[str, Any]]):
    """Return Crunchy API."""
    _api = CrunchyAPI(crunchy_config)
    _api.set_dry_run(True)
    yield _api


@pytest.fixture(name="real_compress_api")
def fixture_real_compress_api(
    demultiplexed_runs: Path, housekeeper_api: HousekeeperAPI, real_crunchy_api: CrunchyAPI
) -> CompressAPI:
    """Return a compress API context."""
    return CompressAPI(
        crunchy_api=real_crunchy_api,
        hk_api=housekeeper_api,
        demux_root=demultiplexed_runs.as_posix(),
    )


@pytest.fixture(name="real_populated_compress_fastq_api")
def fixture_real_populated_compress_fastq_api(
    real_compress_api: CompressAPI, compress_hk_fastq_bundle: dict, helpers: StoreHelpers
) -> CompressAPI:
    """Return real populated compress API."""
    helpers.ensure_hk_bundle(real_compress_api.hk_api, compress_hk_fastq_bundle)

    return real_compress_api


@pytest.fixture(name="real_populated_compress_context")
def fixture_real_populated_compress_context(
    real_populated_compress_fastq_api: CompressAPI,
    populated_compress_store: Store,
    cg_config_object: CGConfig,
) -> CGConfig:
    """Return a compress context populated with a completed analysis."""
    # Make sure that there is a case where analysis is completed
    cg_config_object.meta_apis["compress_api"] = real_populated_compress_fastq_api
    cg_config_object.status_db_ = populated_compress_store
    return cg_config_object



"""Common CLI file generation functions"""

import click

from cg.cli.generate.report.base import delivery_report, available_delivery_reports


@click.group()
def generate():
    """Generates and/or modifies files."""


generate.add_command(delivery_report)
generate.add_command(available_delivery_reports)



import logging
from pathlib import Path

from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.hermes.models import CGDeliverables
from cg.cli.workflow.balsamic.base import balsamic, start, start_available, store, store_available
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.cg_config import CGConfig
from click.testing import CliRunner

EXIT_SUCCESS = 0


def test_balsamic_no_args(cli_runner: CliRunner, balsamic_context: CGConfig):
    """Test to see that running BALSAMIC without options prints help and doesn't result in an error"""
    # GIVEN no arguments or options besides the command call

    # WHEN running command
    result = cli_runner.invoke(balsamic, [], obj=balsamic_context)

    # THEN command runs successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN help should be printed
    assert "help" in result.output


def test_start(cli_runner: CliRunner, balsamic_context: CGConfig, mock_config, caplog, mocker):
    """Test to ensure all parts of start command will run successfully given ideal conditions"""
    caplog.set_level(logging.INFO)

    # GIVEN case id for which we created a config file
    case_id = "balsamic_case_wgs_single"

    # GIVEN decompression is not needed
    mocker.patch.object(BalsamicAnalysisAPI, "resolve_decompression")
    BalsamicAnalysisAPI.resolve_decompression.return_value = None

    # WHEN dry running with dry specified
    result = cli_runner.invoke(start, [case_id, "--dry-run"], obj=balsamic_context)

    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS
    assert case_id in caplog.text


def test_store(
    cli_runner: CliRunner,
    balsamic_context: CGConfig,
    real_housekeeper_api,
    mock_config,
    mock_deliverable,
    mock_analysis_finish,
    caplog,
    hermes_deliverables,
    mocker,
):
    """Test to ensure all parts of store command are run successfully given ideal conditions"""
    caplog.set_level(logging.INFO)

    # GIVEN case-id for which we created a config file, deliverables file, and analysis_finish file
    case_id = "balsamic_case_wgs_single"

    # Set Housekeeper to an empty real Housekeeper store
    balsamic_context.housekeeper_api_ = real_housekeeper_api
    balsamic_context.meta_apis["analysis_api"].housekeeper_api = real_housekeeper_api

    # Make sure the bundle was not present in the store
    assert not balsamic_context.housekeeper_api.bundle(case_id)

    # Make sure  analysis not already stored in ClinicalDB
    assert not balsamic_context.status_db.get_case_by_internal_id(internal_id=case_id).analyses

    # GIVEN that HermesAPI returns a deliverables output
    mocker.patch.object(HermesApi, "convert_deliverables")
    HermesApi.convert_deliverables.return_value = CGDeliverables(**hermes_deliverables)

    # WHEN running command
    result = cli_runner.invoke(store, [case_id, "--dry-run"], obj=balsamic_context)

    # THEN bundle should be successfully added to HK and STATUS
    assert result.exit_code == EXIT_SUCCESS
    assert "Analysis successfully stored in Housekeeper" in caplog.text
    assert "Analysis successfully stored in StatusDB" in caplog.text
    assert balsamic_context.status_db.get_case_by_internal_id(internal_id=case_id).analyses
    assert balsamic_context.housekeeper_api.bundle(case_id)


def test_start_available(cli_runner: CliRunner, balsamic_context: CGConfig, caplog, mocker):
    """Test to ensure all parts of compound start-available command are executed given ideal conditions
    Test that start-available picks up eligible cases and does not pick up ineligible ones"""
    caplog.set_level(logging.INFO)

    # GIVEN CASE ID of sample where read counts pass threshold
    case_id_success = "balsamic_case_wgs_paired_enough_reads"

    # GIVEN CASE ID where read counts did not pass the threshold
    case_id_fail = "balsamic_case_tgs_paired"

    # Ensure the config is mocked to run compound command
    Path.mkdir(
        Path(
            balsamic_context.meta_apis["analysis_api"].get_case_config_path(case_id_success)
        ).parent,
        exist_ok=True,
    )
    Path(balsamic_context.meta_apis["analysis_api"].get_case_config_path(case_id_success)).touch(
        exist_ok=True
    )

    # GIVEN decompression is not needed
    mocker.patch.object(BalsamicAnalysisAPI, "resolve_decompression")
    BalsamicAnalysisAPI.resolve_decompression.return_value = None

    # WHEN running command
    result = cli_runner.invoke(start_available, ["--dry-run"], obj=balsamic_context)

    # THEN command exits with 1 because one of cases raised errors
    assert result.exit_code == 1

    # THEN it should successfully identify the one case eligible for auto-start
    assert case_id_success in caplog.text

    # THEN the ineligible case should NOT be ran
    assert case_id_fail not in caplog.text

    # THEN action of the case should NOT be set to running
    assert balsamic_context.status_db.get_case_by_internal_id(case_id_fail).action is None


def test_store_available(
    tmpdir_factory,
    cli_runner: CliRunner,
    balsamic_context: CGConfig,
    real_housekeeper_api,
    mock_config,
    mock_deliverable,
    mock_analysis_finish,
    caplog,
    mocker,
    hermes_deliverables,
):
    """Test to ensure all parts of compound store-available command are executed given ideal conditions
    Test that sore-available picks up eligible cases and does not pick up ineligible ones"""
    caplog.set_level(logging.INFO)

    # GIVEN CASE ID of sample where read counts pass threshold
    case_id_success = "balsamic_case_wgs_single"

    # GIVEN CASE ID where analysis finish is not mocked
    case_id_fail = "balsamic_case_wgs_paired"

    # Ensure the config is mocked for fail case to run compound command
    Path.mkdir(
        Path(balsamic_context.meta_apis["analysis_api"].get_case_config_path(case_id_fail)).parent,
        exist_ok=True,
    )
    Path(balsamic_context.meta_apis["analysis_api"].get_case_config_path(case_id_fail)).touch(
        exist_ok=True
    )

    # GIVEN that HermesAPI returns a deliverables output
    mocker.patch.object(HermesApi, "convert_deliverables")
    HermesApi.convert_deliverables.return_value = CGDeliverables(**hermes_deliverables)

    # Ensure case was successfully picked up by start-available and status set to running
    result = cli_runner.invoke(start_available, ["--dry-run"], obj=balsamic_context)
    balsamic_context.status_db.get_case_by_internal_id(case_id_success).action = "running"
    balsamic_context.status_db.commit()

    # THEN command exits with 1 because one of the cases threw errors
    assert result.exit_code == 1
    assert case_id_success in caplog.text
    assert balsamic_context.status_db.get_case_by_internal_id(case_id_success).action == "running"

    balsamic_context.housekeeper_api_ = real_housekeeper_api
    balsamic_context.meta_apis["analysis_api"].housekeeper_api = real_housekeeper_api

    # WHEN running command
    result = cli_runner.invoke(store_available, obj=balsamic_context)

    # THEN command exits successfully
    assert result.exit_code == 0

    # THEN case id with analysis_finish gets picked up
    assert case_id_success in caplog.text

    # THEN case has analyses
    assert balsamic_context.status_db.get_case_by_internal_id(case_id_success).analyses

    # THEN bundle can be found in Housekeeper
    assert balsamic_context.housekeeper_api.bundle(case_id_success)

    # THEN bundle added successfully and action set to None
    assert balsamic_context.status_db.get_case_by_internal_id(case_id_success).action is None



from pathlib import Path

from cg.apps.cgstats.parsers.run_info import RunInfo


def test_parse_run_info(run_info_path: Path):
    """Test if RunInfo parser is working properly and generates information"""
    # GIVEN an existing RunInfo.xml path
    assert run_info_path.exists()

    # WHEN instantiation a RunInfo object
    run_info_obj: RunInfo = RunInfo(run_info_path)

    # THEN the object should be successfully parsed and contain information_sheet
    assert run_info_obj.basemask


def test_run_info_basemask(run_info_path: Path):
    """Test if basemask generation is working as expected"""
    # GIVEN a expected basemask structure
    expected_basemask: str = "51,8,8,51"

    # WHEN generating the basemask from runinfo
    generated_basemask: str = RunInfo(run_info_path).basemask

    # THEN the generated basemask should match the expected one
    assert generated_basemask == expected_basemask



"""This script tests the cli methods to set flow cells in status-db."""
from cg.cli.set.base import flowcell
from cg.constants import FLOWCELL_STATUS
from cg.models.cg_config import CGConfig
from cg.store import Store
from click.testing import CliRunner

SUCCESS = 0


def test_set_flowcell_bad_flowcell(cli_runner: CliRunner, base_context: CGConfig):
    """Test to set a flow cell using a non-existing flowcell."""
    # GIVEN an empty database

    # WHEN setting a flow cell
    flow_cell_name = "dummy_name"
    result = cli_runner.invoke(flowcell, [flow_cell_name], obj=base_context)

    # THEN it should complain in missing flow cell instead of setting a flow cell
    assert result.exit_code != SUCCESS


def test_set_flowcell_required(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    """Test to set a flow cell using only the required arguments."""
    # GIVEN a database with a flow cell
    flow_cell_name = helpers.add_flowcell(base_store).name
    assert base_store.Flowcell.query.count() == 1

    # WHEN setting a flowcell
    result = cli_runner.invoke(flowcell, [flow_cell_name], obj=base_context)

    # THEN it should have been set
    assert result.exit_code == SUCCESS


def test_set_flowcell_status(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    """Test that the updated flow_cell get the status we send in."""
    # GIVEN a database with a flow cell
    flow_cell_name = helpers.add_flowcell(base_store).name
    status = FLOWCELL_STATUS[2]
    assert base_store.Flowcell.query.first().status != status

    # WHEN setting a flowcell
    result = cli_runner.invoke(flowcell, ["--status", status, flow_cell_name], obj=base_context)

    # THEN it should have been set
    assert result.exit_code == SUCCESS
    assert base_store.Flowcell.query.count() == 1
    assert base_store.Flowcell.query.first().status == status



"""Module for defining invoice models."""
from pydantic import BaseModel
from typing import List, Optional, Any
from cg.constants.priority import PriorityTerms
from cg.constants.sequencing import RecordType


class InvoiceContact(BaseModel):
    """Class for collection contact information used in the invoice."""

    name: str
    email: str
    customer_name: str
    reference: str
    address: str


class InvoiceApplication(BaseModel):
    """Class to collect Application information used in the invoice."""

    version: str
    tag: str
    discounted_price: int
    percent_kth: int


class InvoiceInfo(BaseModel):
    """Class to collect invoice information used in the invoice report."""

    name: str
    id: str
    lims_id: Optional[str]
    application_tag: str
    project: str
    date: Optional[Any]
    price: int
    priority: PriorityTerms
    price_kth: Optional[int]
    total_price: Optional[int]


class InvoiceReport(BaseModel):
    """Class that collects information used to create the invoice Excel sheet."""

    cost_center: str
    project_number: Optional[str]
    customer_id: str
    customer_name: str
    agreement: Optional[str]
    invoice_id: int
    contact: dict
    records: List
    pooled_samples: Optional[Any]
    record_type: RecordType



"""Tests cli methods to create the case config for RNAfusion"""

import logging
from typing import List

from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.workflow.rnafusion.base import config_case
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


def test_without_options(cli_runner: CliRunner, rnafusion_context: CGConfig):
    """Test command without case_id."""
    # GIVEN NO case_id
    # WHEN dry running without anything specified
    result = cli_runner.invoke(config_case, obj=rnafusion_context)
    # THEN command should mention argument
    assert result.exit_code != EXIT_SUCCESS
    assert "Missing argument" in result.output


def test_with_missing_case(
    cli_runner: CliRunner,
    rnafusion_context: CGConfig,
    caplog: LogCaptureFixture,
    not_existing_case_id: str,
):
    """Test command with invalid case to start with."""
    caplog.set_level(logging.ERROR)
    # GIVEN case_id not in database
    assert not rnafusion_context.status_db.get_case_by_internal_id(internal_id=not_existing_case_id)
    # WHEN running
    result = cli_runner.invoke(config_case, [not_existing_case_id], obj=rnafusion_context)
    # THEN command should NOT successfully call the command it creates
    assert result.exit_code != EXIT_SUCCESS
    # THEN ERROR log should be printed containing invalid case_id
    assert "could not be found in StatusDB!" in caplog.text


def test_without_samples(
    cli_runner: CliRunner,
    rnafusion_context: CGConfig,
    caplog: LogCaptureFixture,
    no_sample_case_id: str,
):
    """Test command with case_id and no samples."""
    caplog.set_level(logging.ERROR)
    # GIVEN case-id
    case_id: str = no_sample_case_id
    # WHEN running config case
    result = cli_runner.invoke(config_case, [case_id], obj=rnafusion_context)
    # THEN command should print the rnafusion command-string
    assert result.exit_code != EXIT_SUCCESS
    # THEN warning should be printed that no sample is found
    assert case_id in caplog.text
    assert "has no samples" in caplog.text


def test_strandedness(
    cli_runner: CliRunner,
    rnafusion_context: CGConfig,
    caplog: LogCaptureFixture,
    rnafusion_case_id: str,
):
    """Test command with --strandedness option."""
    caplog.set_level(logging.INFO)
    # GIVEN a VALID case_id and genome_version
    case_id: str = rnafusion_case_id
    option_key: str = "--strandedness"
    option_values: List[str] = ["reverse", "forward", "unstranded"]
    # WHEN running with strandedness option specified
    for option_value in option_values:
        result = cli_runner.invoke(
            config_case,
            [case_id, option_key, option_value],
            obj=rnafusion_context,
        )
        # THEN command should be generated successfully
        assert result.exit_code == EXIT_SUCCESS
        # THEN dry-print should include the option key and value
        assert option_key.replace("--", "") in caplog.text
        assert option_value in caplog.text


def test_wrong_strandedness(
    cli_runner: CliRunner,
    rnafusion_context: CGConfig,
    caplog: LogCaptureFixture,
    rnafusion_case_id: str,
):
    """Test command with --strandedness option."""
    caplog.set_level(logging.INFO)
    # GIVEN a VALID case_id and genome_version
    case_id: str = rnafusion_case_id
    option_key: str = "--strandedness"
    option_value: str = "unknown"
    # WHEN running with strandedness option specified
    result = cli_runner.invoke(
        config_case,
        [case_id, option_key, option_value],
        obj=rnafusion_context,
    )
    # THEN command should fail
    assert result.exit_code != EXIT_SUCCESS


def test_params_file(
    cli_runner: CliRunner,
    rnafusion_context: CGConfig,
    caplog: LogCaptureFixture,
    rnafusion_case_id: str,
):
    """Test command generates default params_file."""
    caplog.set_level(logging.INFO)

    # GIVEN a VALID case_id and genome_version
    case_id: str = rnafusion_case_id

    # WHEN running config case
    result = cli_runner.invoke(config_case, [case_id], obj=rnafusion_context)

    # THEN command should print the rnafusion command-string
    assert result.exit_code == EXIT_SUCCESS

    # THEN parameters file should be generated
    assert "Generating parameters file" in caplog.text



from sqlalchemy import Column, UniqueConstraint, types

from .base import Model


class Version(Model):
    __table_args__ = (
        UniqueConstraint("name", "major", "minor", "patch", name="name_major_minor_patch_uc"),
    )

    version_id = Column(types.Integer, primary_key=True)
    name = Column(types.String(255))
    major = Column(types.Integer)
    minor = Column(types.Integer)
    patch = Column(types.Integer)
    comment = Column(types.String(255))
    time = Column(types.DateTime, nullable=False)

    @staticmethod
    def check(dbname: str, ver: str) -> bool:
        """Checks version of database against dbname and version

        [normally from the config file]

        Args:
          dbname (str): database name as stored in table version
          ver (str): version string in the format major.minor.patch

        Returns:
          True: if identical
        """
        version: Version = Version.get_version()
        if version is None:
            return False
        ver_string = "{0}.{1}.{2}".format(
            str(version.major), str(version.minor), str(version.patch)
        )
        return (ver_string == ver) and (dbname == version.name)



import datetime as dt
import logging
import shutil
from csv import reader
from pathlib import Path
from typing import List, Optional

import pandas as pd
from alchy import Query
from cg.constants import Pipeline
from cg.constants.demultiplexing import SAMPLE_SHEET_DATA_HEADER
from cg.exc import CgError
from cg.meta.workflow.analysis import AnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Family, Flowcell, Sample
from cg.utils import Process

LOG = logging.getLogger(__name__)


class FluffyAnalysisAPI(AnalysisAPI):
    def __init__(
        self,
        config: CGConfig,
        pipeline: Pipeline = Pipeline.FLUFFY,
    ):
        self.root_dir = Path(config.fluffy.root_dir)
        LOG.info("Set root dir to %s", config.fluffy.root_dir)
        self.fluffy_config = Path(config.fluffy.config_path)
        super().__init__(
            pipeline,
            config,
        )

    @property
    def threshold_reads(self):
        return False

    @property
    def process(self) -> Process:
        if not self._process:
            self._process = Process(binary=self.config.fluffy.binary_path)
        return self._process

    def get_case_path(self, case_id: str) -> Path:
        return Path(self.root_dir, case_id)

    def get_sample_sheet_path(self, case_id: str) -> Path:
        """
        Location in case folder where sample sheet is expected to be stored. Sample sheet is used as a config
        required to run Fluffy
        """
        starlims_id: str = (
            self.status_db.get_case_by_internal_id(internal_id=case_id).links[0].sample.order
        )
        return Path(self.root_dir, case_id, f"SampleSheet_{starlims_id}.csv")

    def get_workdir_path(self, case_id: str) -> Path:
        """
        Location in case folder where all sub-folders for each sample containing fastq files are to e saved
        """
        return Path(self.root_dir, case_id, "fastq")

    def get_fastq_path(self, case_id: str, sample_id: str) -> Path:
        """
        Location in working directory to which fastq files are saved for each sample separately
        """
        return Path(self.get_workdir_path(case_id), sample_id)

    def get_output_path(self, case_id: str) -> Path:
        """
        Location in case directory where Fluffy outputs will be stored
        """
        return Path(self.root_dir, case_id, "output")

    def get_deliverables_file_path(self, case_id: str) -> Path:
        """
        Location in working directory where deliverables file will be stored upon completion of analysis.
        Deliverables file is used to communicate paths and tag definitions for files in a finished analysis
        """
        return Path(self.get_output_path(case_id), "deliverables.yaml")

    def get_trailblazer_config_path(self, case_id: str) -> Path:
        """
        Location in working directory where SLURM job id file is to be stored.
        This file contains SLURM ID of jobs associated with current analysis ,
        is used as a config to be submitted to Trailblazer and track job progress in SLURM
        """
        return Path(self.get_output_path(case_id), "sacct", "submitted_jobs.yaml")

    def get_analysis_finish_path(self, case_id: str) -> Path:
        return Path(self.get_output_path(case_id), "COMPLETE")

    def link_fastq_files(self, case_id: str, dry_run: bool = False) -> None:
        """
        Links fastq files from Housekeeper to case working directory
        """
        case_obj: Family = self.status_db.get_case_by_internal_id(internal_id=case_id)
        latest_flow_cell = self.status_db.get_latest_flow_cell_on_case(family_id=case_id)
        workdir_path = self.get_workdir_path(case_id=case_id)
        if workdir_path.exists() and not dry_run:
            LOG.info("Fastq directory exists, removing and re-linking files!")
            shutil.rmtree(workdir_path, ignore_errors=True)
        workdir_path.mkdir(parents=True, exist_ok=True)
        for family_sample in case_obj.links:
            sample_id = family_sample.sample.internal_id
            files: Query = self.housekeeper_api.files(
                bundle=sample_id, tags=["fastq", latest_flow_cell.name]
            )

            sample_path: Path = self.get_fastq_path(case_id=case_id, sample_id=sample_id)
            for file in files:
                if not dry_run:
                    Path.mkdir(sample_path, exist_ok=True, parents=True)
                    Path(sample_path / Path(file.full_path).name).symlink_to(file.full_path)
                LOG.info(f"Linking {file.full_path} to {sample_path / Path(file.full_path).name}")

    def get_concentrations_from_lims(self, sample_id: str) -> str:
        """Get sample concentration from LIMS"""
        return self.lims_api.get_sample_attribute(lims_id=sample_id, key="concentration_sample")

    def get_sample_starlims_id(self, sample_id: str) -> str:
        sample_obj: Sample = self.status_db.get_sample_by_internal_id(sample_id)
        return sample_obj.order

    def get_sample_sequenced_date(self, sample_id: str) -> Optional[dt.date]:
        sample_obj: Sample = self.status_db.get_sample_by_internal_id(sample_id)
        sequenced_at: dt.datetime = sample_obj.sequenced_at
        if sequenced_at:
            return sequenced_at.date()

    def get_sample_control_status(self, sample_id: str) -> bool:
        sample_obj: Sample = self.status_db.get_sample_by_internal_id(sample_id)
        return bool(sample_obj.control)

    def get_nr_of_header_lines_in_sample_sheet(
        self,
        sample_sheet_housekeeper_path: Path,
    ) -> int:
        """
        Return the number of header lines in a Fluffy sample sheet.
        Any lines before and including the line starting with [Data] is considered the header.

        Returns:
        int: The number of lines before the SAMPLE_SHEET_DATA_HEADER
        """
        with sample_sheet_housekeeper_path.open("r") as read_obj:
            csv_reader = reader(read_obj)
            header_line_count: int = 1
            for line in csv_reader:
                if SAMPLE_SHEET_DATA_HEADER in line:
                    break
                header_line_count += 1
        return header_line_count

    def read_sample_sheet_data(self, sample_sheet_housekeeper_path: Path) -> pd.DataFrame:
        """
        Read in a sample sheet starting from the SAMPLE_SHEET_DATA_HEADER.

        Args:
                        sample_sheet_housekeeper_path (Path): Path to the housekeeper sample sheet file

        Returns:
                        pd.DataFrame: A pandas dataframe of the sample sheet
        """
        header_line_count: int = self.get_nr_of_header_lines_in_sample_sheet(
            sample_sheet_housekeeper_path=sample_sheet_housekeeper_path
        )
        return pd.read_csv(sample_sheet_housekeeper_path, index_col=None, header=header_line_count)

    def add_sample_sheet_column(
        self, sample_sheet_df: pd.DataFrame, new_column: str, to_add: list
    ) -> pd.DataFrame:
        """Add columns to the sample sheet.
        Returns:
                        pd.DataFrame: Sample sheet pd.DataFrame.
        """
        try:
            sample_sheet_df[new_column] = to_add
        except ValueError:
            LOG.error(
                f"Error when trying to add the column: {new_column} to sample sheet with data: {to_add}."
            )
        return sample_sheet_df

    def column_has_alias(self, sample_sheet_df: pd.DataFrame, alias: str) -> bool:
        return alias in sample_sheet_df.columns

    def set_column_alias(self, sample_sheet_df: pd.DataFrame, alias: str, alternative: str) -> str:
        """Return column alias from the sample sheet or set to alternative.
        Returns: str: column name alias
        """
        return (
            alias
            if self.column_has_alias(sample_sheet_df=sample_sheet_df, alias=alias)
            else alternative
        )

    def write_sample_sheet_csv(
        self, sample_sheet_df: pd.DataFrame, sample_sheet_workdir_path: Path
    ):
        """
        Write the sample sheet as a csv file
        """
        sample_sheet_df.to_csv(
            sample_sheet_workdir_path,
            sep=",",
            index=False,
        )

    def populate_sample_sheet(
        self, sample_sheet_housekeeper_path: Path, sample_sheet_workdir_path: Path
    ) -> None:
        """
        Reads the fluffy samplesheet *.csv file as found in Housekeeper.
        Edits column 'SampleName' to include customer name for sample.
        Edits column 'Sample_Project or Project' to include customer sample starlims id.
        Adds columns Library_nM, SequencingDate, Exclude and populates with orderform values
        """
        sample_sheet_df = self.read_sample_sheet_data(
            sample_sheet_housekeeper_path=sample_sheet_housekeeper_path
        )

        sample_id_column_alias = self.set_column_alias(
            sample_sheet_df=sample_sheet_df, alias="Sample_ID", alternative="SampleID"
        )

        sample_project_column_alias = self.set_column_alias(
            sample_sheet_df=sample_sheet_df, alias="Sample_Project", alternative="Project"
        )

        column_to_value_map: dict = {
            "Exclude": lambda x: self.get_sample_control_status(sample_id=x),
            "SampleName": lambda x: self.get_sample_name_from_lims_id(lims_id=x),
            "Library_nM": lambda x: self.get_concentrations_from_lims(sample_id=x),
            "SequencingDate": lambda x: self.get_sample_sequenced_date(sample_id=x),
            sample_project_column_alias: lambda x: self.get_sample_starlims_id(sample_id=x),
        }

        for column, value in column_to_value_map.items():
            sample_sheet_df = self.add_sample_sheet_column(
                sample_sheet_df=sample_sheet_df,
                new_column=column,
                to_add=sample_sheet_df[sample_id_column_alias].apply(value),
            )

        self.write_sample_sheet_csv(
            sample_sheet_df=sample_sheet_df, sample_sheet_workdir_path=sample_sheet_workdir_path
        )

    def get_sample_sheet_housekeeper_path(self, flowcell_name: str) -> Path:
        """
        Returns the path to original samplesheet file that is added to Housekeeper
        """
        sample_sheet_query: list = self.housekeeper_api.files(
            bundle=flowcell_name, tags=["samplesheet"]
        ).all()
        if not sample_sheet_query:
            LOG.error(
                "Samplesheet file for flowcell %s could not be found in Housekeeper!", flowcell_name
            )
            raise CgError
        return Path(sample_sheet_query[0].full_path)

    def make_sample_sheet(self, case_id: str, dry_run: bool) -> None:
        """
        Create SampleSheet.csv file in working directory and add desired values to the file
        """
        latest_flow_cell: Flowcell = self.status_db.get_latest_flow_cell_on_case(family_id=case_id)
        sample_sheet_housekeeper_path = self.get_sample_sheet_housekeeper_path(
            flowcell_name=latest_flow_cell.name
        )
        sample_sheet_workdir_path = Path(self.get_sample_sheet_path(case_id=case_id))
        LOG.info(
            "Writing modified csv from %s to %s",
            sample_sheet_housekeeper_path,
            sample_sheet_workdir_path,
        )
        if not dry_run:
            Path(self.root_dir, case_id).mkdir(parents=True, exist_ok=True)
            self.populate_sample_sheet(
                sample_sheet_housekeeper_path=sample_sheet_housekeeper_path,
                sample_sheet_workdir_path=sample_sheet_workdir_path,
            )

    def run_fluffy(
        self, case_id: str, dry_run: bool, workflow_config: str, external_ref: bool = False
    ) -> None:
        """
        Call fluffy with the configured command-line arguments
        """
        output_path: Path = self.get_output_path(case_id=case_id)
        if output_path.exists():
            LOG.info("Old working directory found, cleaning!")
            if not dry_run:
                shutil.rmtree(output_path, ignore_errors=True)
        if not workflow_config:
            workflow_config = self.fluffy_config.as_posix()
        if not external_ref:
            batch_ref_flag = "--batch-ref"
        else:
            batch_ref_flag = ""
        command_args = [
            "--config",
            workflow_config,
            "--sample",
            self.get_sample_sheet_path(case_id=case_id).as_posix(),
            "--project",
            self.get_workdir_path(case_id=case_id).as_posix(),
            "--out",
            self.get_output_path(case_id=case_id).as_posix(),
            "--analyse",
            batch_ref_flag,
            "--slurm_params",
            self.get_slurm_param_qos(case_id=case_id),
        ]
        self.process.run_command(command_args, dry_run=dry_run)

    def get_cases_to_store(self) -> List[Family]:
        """Retrieve a list of cases where analysis finished successfully,
        and is ready to be stored in Housekeeper"""
        return [
            case_object
            for case_object in self.get_running_cases()
            if Path(self.get_analysis_finish_path(case_id=case_object.internal_id)).exists()
        ]

    def get_slurm_param_qos(self, case_id):
        return f"qos:{self.get_slurm_qos_for_case(case_id=case_id)}"



import datetime as dt
from typing import List

from cgmodels.cg.constants import Pipeline

from cg.constants import DataDelivery, GenePanelMasterList
from cg.constants.constants import PrepCategory
from cg.constants.invoice import CustomerNames
from cg.exc import OrderError
from cg.meta.orders.lims import process_lims
from cg.meta.orders.submitter import Submitter
from cg.models.orders.order import OrderIn
from cg.models.orders.sample_base import StatusEnum
from cg.constants.priority import Priority
from cg.store.models import Sample, Family, FamilySample, Customer, ApplicationVersion


class FastqSubmitter(Submitter):
    def submit_order(self, order: OrderIn) -> dict:
        """Submit a batch of samples for FASTQ delivery."""

        project_data, lims_map = process_lims(
            lims_api=self.lims, lims_order=order, new_samples=order.samples
        )
        status_data = self.order_to_status(order)
        self._fill_in_sample_ids(samples=status_data["samples"], lims_map=lims_map)
        new_samples = self.store_items_in_status(
            customer_id=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"],
            ticket_id=order.ticket,
            items=status_data["samples"],
        )
        self._add_missing_reads(new_samples)
        return {"project": project_data, "records": new_samples}

    @staticmethod
    def order_to_status(order: OrderIn) -> dict:
        """Convert order input to status for fastq-only orders."""
        status_data = {
            "customer": order.customer,
            "order": order.name,
            "samples": [
                {
                    "application": sample.application,
                    "capture_kit": sample.capture_kit,
                    "comment": sample.comment,
                    "data_analysis": sample.data_analysis,
                    "data_delivery": sample.data_delivery,
                    "name": sample.name,
                    "priority": sample.priority,
                    "sex": sample.sex,
                    "tumour": sample.tumour,
                    "volume": sample.volume,
                }
                for sample in order.samples
            ],
        }
        return status_data

    def create_maf_case(self, sample_obj: Sample) -> None:
        """Add a MAF case to the Status database."""
        case: Family = self.status.add_case(
            data_analysis=Pipeline(Pipeline.MIP_DNA),
            data_delivery=DataDelivery(DataDelivery.NO_DELIVERY),
            name="_".join([sample_obj.name, "MAF"]),
            panels=[GenePanelMasterList.OMIM_AUTO],
            priority=Priority.research,
            ticket=sample_obj.original_ticket,
        )
        case.customer: Customer = self.status.get_customer_by_internal_id(
            customer_internal_id=CustomerNames.CG_INTERNAL_CUSTOMER
        )
        relationship: FamilySample = self.status.relate_sample(
            family=case, sample=sample_obj, status=StatusEnum.unknown
        )
        self.status.add(case, relationship)

    def store_items_in_status(
        self, customer_id: str, order: str, ordered: dt.datetime, ticket_id: str, items: List[dict]
    ) -> List[Sample]:
        """Store fastq samples in the status database including family connection and delivery"""
        customer: Customer = self.status.get_customer_by_internal_id(
            customer_internal_id=customer_id
        )
        if not customer:
            raise OrderError(f"Unknown customer: {customer_id}")
        new_samples = []
        case: Family = self.status.get_case_by_name_and_customer(
            customer=customer, case_name=ticket_id
        )
        submitted_case: dict = items[0]
        with self.status.session.no_autoflush:
            for sample in items:
                new_sample = self.status.add_sample(
                    name=sample["name"],
                    sex=sample["sex"] or "unknown",
                    comment=sample["comment"],
                    internal_id=sample.get("internal_id"),
                    order=order,
                    ordered=ordered,
                    original_ticket=ticket_id,
                    priority=sample["priority"],
                    tumour=sample["tumour"],
                    capture_kit=sample["capture_kit"],
                )
                new_sample.customer: Customer = customer
                application_tag: str = sample["application"]
                application_version: ApplicationVersion = (
                    self.status.get_current_application_version_by_tag(tag=application_tag)
                )
                if application_version is None:
                    raise OrderError(f"Invalid application: {sample['application']}")
                new_sample.application_version: ApplicationVersion = application_version
                new_samples.append(new_sample)
                if not case:
                    case = self.status.add_case(
                        data_analysis=Pipeline(submitted_case["data_analysis"]),
                        data_delivery=DataDelivery(submitted_case["data_delivery"]),
                        name=ticket_id,
                        panels=None,
                        priority=submitted_case["priority"],
                        ticket=ticket_id,
                    )
                if (
                    not new_sample.is_tumour
                    and new_sample.application_version.application.prep_category
                    == PrepCategory.WHOLE_GENOME_SEQUENCING
                ):
                    self.create_maf_case(sample_obj=new_sample)
                case.customer = customer
                new_relationship = self.status.relate_sample(
                    family=case, sample=new_sample, status=StatusEnum.unknown
                )
                new_delivery = self.status.add_delivery(destination="caesar", sample=new_sample)
                self.status.add(case, new_relationship, new_delivery)

        self.status.add_commit(new_samples)
        return new_samples



"""Test how the api handles files."""
from datetime import datetime
from pathlib import Path
from typing import List, Dict, Any

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from tests.meta.compress.conftest import MockCompressionData
from tests.mocks.hk_mock import MockHousekeeperAPI

from housekeeper.store.models import Version, File

from tests.small_helpers import SmallHelpers
from tests.store_helpers import StoreHelpers


def test_new_file(bed_file: Path, housekeeper_api: MockHousekeeperAPI, small_helpers: SmallHelpers):
    """Test to create a new file with the Housekeeper API"""
    # GIVEN a housekeeper api without files and the path to an existing file
    assert small_helpers.length_of_iterable(housekeeper_api.files()) == 0
    assert bed_file.exists() is True

    # WHEN creating a new file
    new_file = housekeeper_api.new_file(path=bed_file.as_posix())

    # THEN assert a file object was created
    assert new_file

    # THEN assert that the path is correct
    assert new_file.path == bed_file.as_posix()

    # THEN assert that no file is added to the database
    assert small_helpers.length_of_iterable(housekeeper_api.files()) == 0


def test_new_file_non_existing_path(housekeeper_api: MockHousekeeperAPI):
    """Test to create a new file with the Housekeeper API."""
    # GIVEN a housekeeper api without files and the path to a not existing file
    file_name = Path("a_file.hello")
    assert file_name.exists() is False

    # WHEN creating a new file
    new_file: File = housekeeper_api.new_file(path=file_name.as_posix())

    # THEN assert a file object was created
    assert new_file

    # THEN assert that the path is correct
    assert new_file.path == file_name.as_posix()


def test_add_new_file(
    populated_housekeeper_api: MockHousekeeperAPI,
    case_id: str,
    madeline_output: Path,
    small_helpers: SmallHelpers,
    not_existing_hk_tag: str,
):
    """Test to create a new file with the Housekeeper API."""
    # GIVEN a populated housekeeper api and the existing version of a bundle
    version: Version = populated_housekeeper_api.last_version(bundle=case_id)

    # GIVEN an existing file that is not included in the database
    assert madeline_output.exists() is True

    # GIVEN a tag that does not exist
    assert populated_housekeeper_api.get_tag(name=not_existing_hk_tag) is None

    # GIVEN a known number of files in the db
    nr_files_in_db = small_helpers.length_of_iterable(populated_housekeeper_api.files())

    # WHEN creating a new file
    new_file: File = populated_housekeeper_api.add_file(
        path=madeline_output, version_obj=version, tags=not_existing_hk_tag
    )

    # THEN assert a file object was created
    assert new_file

    # THEN assert that the path is correct
    assert new_file.path == madeline_output.resolve().as_posix()

    # THEN assert that the file was not added to the database
    new_nr_files = small_helpers.length_of_iterable(populated_housekeeper_api.files())
    assert new_nr_files == nr_files_in_db + 1


def test_get_files(
    populated_housekeeper_api: MockHousekeeperAPI,
    case_id: str,
    tags: List[str],
    small_helpers: SmallHelpers,
):
    """Test get files method."""
    # GIVEN a populated housekeeper api with some files
    nr_files = small_helpers.length_of_iterable(populated_housekeeper_api.files())
    assert nr_files > 0

    # WHEN fetching all files
    files = populated_housekeeper_api.get_files(bundle=case_id, tags=tags)

    # THEN assert all files where fetched
    assert small_helpers.length_of_iterable(files) == nr_files


def test_get_file(populated_housekeeper_api: MockHousekeeperAPI):
    """Test to get a file from the database."""
    # GIVEN a housekeeper api with a file
    hk_file: File = populated_housekeeper_api.files().first()

    # GIVEN the id of a file that exists in HK
    assert hk_file

    # WHEN fetching the file with get_file
    hk_file = populated_housekeeper_api.get_file(hk_file.id)

    # THEN assert a file was returned
    assert hk_file is not None


def test_get_files_from_version(
    helpers: StoreHelpers,
    real_housekeeper_api: HousekeeperAPI,
    hk_bundle_data: Dict[str, Any],
    hk_tag: str,
    observations_clinical_snv_file_path: Path,
    observations_clinical_sv_file_path: Path,
):
    """Test get the files from the Housekeeper database given the version object."""

    # GIVEN a Housekeeper API with some files
    version: Version = helpers.ensure_hk_version(real_housekeeper_api, hk_bundle_data)
    first_file: File = real_housekeeper_api.add_file(
        path=observations_clinical_snv_file_path, version_obj=version, tags=hk_tag
    )
    second_file: File = real_housekeeper_api.add_file(
        path=observations_clinical_sv_file_path, version_obj=version, tags=hk_tag
    )

    # GIVEN that the files exist in the version object
    assert first_file in version.files
    assert second_file in version.files

    # WHEN extracting the files from version
    files: List[File] = real_housekeeper_api.get_files_from_version(version=version, tags={hk_tag})

    # THEN the added files should be retrieved
    assert first_file in files
    assert second_file in files


def test_get_latest_file_from_version(
    helpers: StoreHelpers,
    real_housekeeper_api: HousekeeperAPI,
    hk_bundle_data: Dict[str, Any],
    hk_tag: str,
    observations_clinical_snv_file_path: Path,
    observations_clinical_sv_file_path: Path,
):
    """Test to get the latest file from the Housekeeper database given the version object."""

    # GIVEN a Housekeeper API with some files
    version: Version = helpers.ensure_hk_version(real_housekeeper_api, hk_bundle_data)
    first_file: File = real_housekeeper_api.add_file(
        path=observations_clinical_snv_file_path, version_obj=version, tags=hk_tag
    )
    second_file: File = real_housekeeper_api.add_file(
        path=observations_clinical_sv_file_path, version_obj=version, tags=hk_tag
    )

    # GIVEN that the files exist in the version object
    assert first_file in version.files
    assert second_file in version.files

    # WHEN extracting the latest file from version
    latest_file: File = real_housekeeper_api.get_latest_file_from_version(
        version=version, tags={hk_tag}
    )

    # THEN the file with the higher ID should be returned
    assert latest_file == second_file


def test_get_file_from_latest_version(case_id: str, populated_housekeeper_api: MockHousekeeperAPI):
    """Test to get a file from the database from the latest version."""
    # GIVEN a housekeeper api with a file
    hk_file: File = populated_housekeeper_api.files().first()

    # GIVEN a tag of a file that exists in HK
    assert hk_file.tags

    # WHEN fetching the file
    hk_file: File = populated_housekeeper_api.get_file_from_latest_version(
        bundle_name=case_id, tags=[hk_file.tags[0].name]
    )

    # THEN assert a file was returned
    assert hk_file is not None


def test_get_files_from_latest_version(
    case_id: str, populated_housekeeper_api: MockHousekeeperAPI, small_helpers: SmallHelpers
):
    """Test to get files from the database from the latest version."""

    # GIVEN a Housekeeper version
    version: Version = populated_housekeeper_api.last_version(bundle=case_id)

    # GIVEN a housekeeper api with a file
    hk_file: File = populated_housekeeper_api.files().first()

    # GIVEN a tag of a file that exists in HK
    assert hk_file.tags

    # GIVEN another file with the same tag
    populated_housekeeper_api.add_file(
        path=Path("a_new_file.bed").as_posix(), tags=[hk_file.tags[0].name], version_obj=version
    )

    # WHEN fetching the files
    hk_files: List[File] = populated_housekeeper_api.get_files_from_latest_version(
        bundle_name=case_id, tags=[hk_file.tags[0].name]
    )

    # THEN assert 2 files were returned
    assert small_helpers.length_of_iterable(hk_files) == 2


def test_delete_file(populated_housekeeper_api: HousekeeperAPI):
    """Test to delete a file from the database."""
    # GIVEN a housekeeper api with a file
    hk_file: File = populated_housekeeper_api.files().first()

    # GIVEN the id of a file that exists in HK
    assert hk_file

    # WHEN deleting the file
    populated_housekeeper_api.delete_file(hk_file.id)

    # THEN assert the file was removed
    assert populated_housekeeper_api.get_file(hk_file.id) is None


def test_get_included_path(populated_housekeeper_api: MockHousekeeperAPI, case_id: str):
    """Test to get the included path for a file."""
    # GIVEN a populated housekeeper api and the root dir
    root_dir: Path = Path(populated_housekeeper_api.get_root_dir())

    # GIVEN a version and a file object
    version: Version = populated_housekeeper_api.last_version(case_id)
    hk_file: File = version.files[0]

    # WHEN fetching the included path
    included_path: Path = populated_housekeeper_api.get_included_path(
        root_dir=root_dir, version_obj=version, file_obj=hk_file
    )

    # THEN assert that there is no file existing in the included path
    assert included_path.exists() is False

    # THEN assert that the correct path was created
    assert included_path == Path(root_dir, version.relative_root_dir, Path(hk_file.path).name)


def test_get_include_file(populated_housekeeper_api: MockHousekeeperAPI, case_id: str):
    """Test to get the included path for a file."""
    # GIVEN a populated housekeeper api and the root dir
    root_dir: Path = Path(populated_housekeeper_api.get_root_dir())
    version: Version = populated_housekeeper_api.last_version(case_id)
    hk_file: File = version.files[0]
    original_path: Path = Path(hk_file.path)
    included_path: Path = Path(root_dir, version.relative_root_dir, original_path.name)

    # GIVEN that the included file does not exist
    assert included_path.exists() is False

    # WHEN including the file
    included_file = populated_housekeeper_api.include_file(hk_file, version)

    # THEN assert that the file has been linked to the included place
    assert included_path.exists() is True

    # THEN assert that the file path has been updated
    assert included_file.path != original_path


def test_include_files_to_latest_version_when_included(
    caplog, case_id: str, populated_housekeeper_api: MockHousekeeperAPI
):
    """Test to include files for a bundle."""
    # GIVEN a populated Housekeeper API and the root dir
    root_dir: Path = Path(populated_housekeeper_api.get_root_dir())
    version: Version = populated_housekeeper_api.last_version(case_id)
    hk_file: File = version.files[0]
    original_path: Path = Path(hk_file.path)
    included_dir_path: Path = Path(root_dir, version.relative_root_dir)
    included_dir_path.mkdir(parents=True, exist_ok=True)
    included_path: Path = Path(included_dir_path, original_path.name)
    included_path.touch()

    # GIVEN that the included file does exist
    assert included_path.exists() is True

    # WHEN including the file
    populated_housekeeper_api.include_files_to_latest_version(bundle_name=case_id)

    hk_version: Version = populated_housekeeper_api.get_latest_bundle_version(bundle_name=case_id)
    included_file: File = hk_version.files[0]

    # THEN assert that the file is still linked to the included place
    assert included_path.exists() is True

    # THEN assert that the file path is unchanged
    assert included_file.path == original_path.as_posix()

    assert f"File is already included in Housekeeper for bundle: {case_id}" in caplog.text


def test_include_files_to_latest_version(
    case_id: str,
    madeline_output: Path,
    not_existing_hk_tag: str,
    populated_housekeeper_api: MockHousekeeperAPI,
):
    """Test to include files for a bundle."""
    # GIVEN a populated Housekeeper API and the root dir
    root_dir: Path = Path(populated_housekeeper_api.get_root_dir())
    version: Version = populated_housekeeper_api.last_version(case_id)
    hk_file: File = version.files[0]
    original_path: Path = Path(hk_file.path)
    included_path: Path = Path(root_dir, version.relative_root_dir, original_path.name)

    # GIVEN that the included file does not exist
    assert included_path.exists() is False

    new_file: File = populated_housekeeper_api.add_file(
        path=madeline_output, version_obj=version, tags=not_existing_hk_tag
    )
    populated_housekeeper_api.commit()

    assert Path(new_file.path).parent != included_path.parent

    # WHEN including the file
    populated_housekeeper_api.include_files_to_latest_version(bundle_name=case_id)

    hk_version: Version = populated_housekeeper_api.get_latest_bundle_version(bundle_name=case_id)

    # THEN all files in bundle should be included
    for file in hk_version.files:
        assert file.is_included

    # THEN assert that the file path has been updated
    included_madelaine_file: File = populated_housekeeper_api.get_file(file_id=new_file.id)
    assert Path(included_madelaine_file.full_path).parent == included_path.parent


def test_check_bundle_files(
    case_id: str,
    timestamp: datetime,
    populated_housekeeper_api: MockHousekeeperAPI,
    hk_version: Version,
    fastq_file: Path,
    sample_id: str,
    bed_file: Path,
):
    """Test to see if the function correctly identifies a file that is present and returns a lis without it."""
    # GIVEN a housekeeper version with a file
    version: Version = populated_housekeeper_api.version(bundle=case_id, date=timestamp)

    # WHEN attempting to add two files, one existing and one new
    files_to_add: List[Path] = populated_housekeeper_api.check_bundle_files(
        file_paths=[bed_file, fastq_file],
        bundle_name=case_id,
        last_version=version,
    )

    # Then only the new file should be returned
    assert files_to_add == [fastq_file]


def test_get_tag_names_from_file(populated_housekeeper_api: MockHousekeeperAPI):
    """Test get tag names on a file."""
    # GIVEN a housekeeper api with a file
    hk_file = populated_housekeeper_api.files().first()
    assert hk_file.tags

    # WHEN fetching tags of a file
    tag_names = populated_housekeeper_api.get_tag_names_from_file(hk_file)

    # THEN a list of tag names is returned
    assert tag_names is not None

    # THEN the return type is a list of strings
    assert isinstance(tag_names, list)
    assert all(isinstance(elem, str) for elem in tag_names)


def test_is_fastq_or_spring_in_all_bundles_when_none(
    populated_housekeeper_api: MockHousekeeperAPI,
    case_id: str,
    tags: List[str],
):
    """Test checking if all FASTQ or SPRING files are present in bundles when no files are present."""
    # GIVEN a populated housekeeper api with some files

    # WHEN fetching all files
    was_true = populated_housekeeper_api.is_fastq_or_spring_in_all_bundles(bundle_names=[case_id])

    # THEN assert all file were not present in all bundles
    assert not was_true


def test_is_fastq_or_spring_in_all_bundles(
    populated_housekeeper_api: MockHousekeeperAPI,
    case_id: str,
    madeline_output: Path,
    tags: List[str],
):
    """Test checking if all FASTQ or SPRING files are present in bundles when files are present."""
    # GIVEN a populated housekeeper api with some files

    # GIVEN a FASTQ file tag with a file included the bundle
    populated_housekeeper_api.add_and_include_file_to_latest_version(
        file=madeline_output, bundle_name=case_id, tags=[SequencingFileTag.FASTQ]
    )

    # WHEN fetching all files
    was_true = populated_housekeeper_api.is_fastq_or_spring_in_all_bundles(bundle_names=[case_id])

    # THEN assert all file were present in all bundles
    assert was_true


def test_is_fastq_or_spring_in_all_bundles_when_missing(
    populated_housekeeper_api: MockHousekeeperAPI,
    case_id: str,
    sample_id: str,
    madeline_output: Path,
    tags: List[str],
):
    """Test checking if all FASTQ or SPRING files are present in bundles when not all bundles have files present."""
    # GIVEN a populated housekeeper api with some files
    version: Version = populated_housekeeper_api.last_version(case_id)

    # GIVEN a FASTQ file tag in the bundle
    populated_housekeeper_api.add_file(
        path=madeline_output, version_obj=version, tags=[SequencingFileTag.FASTQ]
    )

    # GIVEN an empty bundle
    populated_housekeeper_api.create_new_bundle_and_version(name=sample_id)

    populated_housekeeper_api.commit()

    # WHEN fetching all files
    was_true = populated_housekeeper_api.is_fastq_or_spring_in_all_bundles(
        bundle_names=[case_id, sample_id]
    )

    # THEN assert all file were not present in all bundles
    assert not was_true


def test_is_fastq_or_spring_in_all_bundles_when_multiple_bundles(
    case_id: str,
    compression_object: MockCompressionData,
    populated_housekeeper_api: MockHousekeeperAPI,
    madeline_output: Path,
    sample_id: str,
    tags: List[str],
):
    """Test checking if all FASTQ or SPRING files are present in bundles when all bundles have files present."""
    # GIVEN a populated housekeeper api with some files

    # GIVEN a FASTQ file tag with a file included the bundle
    populated_housekeeper_api.add_and_include_file_to_latest_version(
        file=madeline_output, bundle_name=case_id, tags=[SequencingFileTag.FASTQ]
    )

    # GIVEN an empty bundle
    populated_housekeeper_api.create_new_bundle_and_version(name=sample_id)

    # GIVEN an existing SPRING metadata file
    compression_object.spring_metadata_path.touch()

    # GIVEN a SPRING file tag with a file included the bundle
    populated_housekeeper_api.add_and_include_file_to_latest_version(
        file=compression_object.spring_metadata_path,
        bundle_name=sample_id,
        tags=[SequencingFileTag.SPRING_METADATA],
    )

    # WHEN fetching all files
    was_true = populated_housekeeper_api.is_fastq_or_spring_in_all_bundles(
        bundle_names=[case_id, sample_id]
    )

    # THEN assert all file were present in all bundles
    assert was_true


def test_is_fastq_or_spring_in_all_bundles_when_multiple_bundles_and_files(
    case_id: str,
    compression_object: MockCompressionData,
    populated_housekeeper_api: MockHousekeeperAPI,
    madeline_output: Path,
    sample_id: str,
    tags: List[str],
):
    """
    Test checking if all FASTQ or SPRING files are present in bundles when all bundles have files present and some
    both FASTQ and SPRING.
    """
    # GIVEN a populated Housekeeper API with some files

    # GIVEN a FASTQ file tag with a file included in the bundle
    populated_housekeeper_api.add_and_include_file_to_latest_version(
        file=madeline_output, bundle_name=case_id, tags=[SequencingFileTag.FASTQ]
    )

    # GIVEN an empty bundle
    populated_housekeeper_api.create_new_bundle_and_version(name=sample_id)

    # GIVEN an existing SPRING metadata file
    compression_object.spring_metadata_path.touch()

    # GIVEN a SPRING file tag with a file included in the bundle
    populated_housekeeper_api.add_and_include_file_to_latest_version(
        file=compression_object.spring_metadata_path,
        bundle_name=sample_id,
        tags=[SequencingFileTag.SPRING_METADATA],
    )

    # WHEN fetching all files
    was_true = populated_housekeeper_api.is_fastq_or_spring_in_all_bundles(
        bundle_names=[case_id, sample_id]
    )

    # THEN assert all file were present in all bundles
    assert was_true


def test_is_fastq_or_spring_on_disk_in_all_bundles_when_none(
    populated_housekeeper_api: MockHousekeeperAPI,
    case_id: str,
    tags: List[str],
):
    """Test checking if all FASTQ or SPRING files are on disk in bundles when no files are on disk."""
    # GIVEN a populated housekeeper api with some files

    # WHEN fetching all files
    was_true = populated_housekeeper_api.is_fastq_or_spring_on_disk_in_all_bundles(
        bundle_names=[case_id]
    )

    # THEN assert all file were not on disk in all bundles
    assert not was_true


def test_is_fastq_or_spring_on_disk_in_all_bundles(
    populated_housekeeper_api: MockHousekeeperAPI,
    case_id: str,
    madeline_output: Path,
    tags: List[str],
):
    """Test checking if all FASTQ or SPRING files are on disk in bundles when files are on disk."""
    # GIVEN a populated housekeeper api with some files

    # GIVEN a FASTQ file tag with a file on disk the bundle
    populated_housekeeper_api.add_and_include_file_to_latest_version(
        file=madeline_output, bundle_name=case_id, tags=[SequencingFileTag.FASTQ]
    )

    # WHEN fetching all files
    was_true = populated_housekeeper_api.is_fastq_or_spring_on_disk_in_all_bundles(
        bundle_names=[case_id]
    )

    # THEN assert all file were on disk in all bundles
    assert was_true


def test_is_fastq_or_spring_on_disk_in_all_bundles_when_missing_file(
    populated_housekeeper_api: MockHousekeeperAPI,
    case_id: str,
    sample_id: str,
    madeline_output: Path,
    tags: List[str],
):
    """Test checking if all FASTQ or SPRING files are on disk in bundles when not all bundles have files on disk."""
    # GIVEN a populated housekeeper api with some files
    version: Version = populated_housekeeper_api.last_version(case_id)

    # GIVEN a FASTQ file tag in the bundle
    populated_housekeeper_api.add_file(
        path=madeline_output, version_obj=version, tags=[SequencingFileTag.FASTQ]
    )

    # GIVEN an empty bundle
    populated_housekeeper_api.create_new_bundle_and_version(name=sample_id)

    # GIVEN a FASTQ file tag in the bundle, but not on disk
    populated_housekeeper_api.add_file(
        path="does_not_exist_on_disk", version_obj=version, tags=[SequencingFileTag.FASTQ]
    )

    populated_housekeeper_api.commit()

    # WHEN fetching all files
    was_true = populated_housekeeper_api.is_fastq_or_spring_on_disk_in_all_bundles(
        bundle_names=[case_id, sample_id]
    )

    # THEN assert all file were not on disk in all bundles
    assert not was_true


def testis_fastq_or_spring_on_disk_in_all_bundles_when_multiple_bundles(
    case_id: str,
    compression_object: MockCompressionData,
    populated_housekeeper_api: MockHousekeeperAPI,
    madeline_output: Path,
    sample_id: str,
    tags: List[str],
):
    """Test checking if all FASTQ or SPRING files are on disk in bundles when all bundles have files on disk."""
    # GIVEN a populated housekeeper api with some files

    # GIVEN a FASTQ file tag with a file included the bundle
    populated_housekeeper_api.add_and_include_file_to_latest_version(
        file=madeline_output, bundle_name=case_id, tags=[SequencingFileTag.FASTQ]
    )

    # GIVEN an empty bundle
    populated_housekeeper_api.create_new_bundle_and_version(name=sample_id)

    # GIVEN an existing SPRING metadata file
    compression_object.spring_metadata_path.touch()

    # GIVEN a SPRING file tag with a file included the bundle
    populated_housekeeper_api.add_and_include_file_to_latest_version(
        file=compression_object.spring_metadata_path,
        bundle_name=sample_id,
        tags=[SequencingFileTag.SPRING_METADATA],
    )

    # WHEN fetching all files
    was_true = populated_housekeeper_api.is_fastq_or_spring_on_disk_in_all_bundles(
        bundle_names=[case_id, sample_id]
    )

    # THEN assert all file were on disk in all bundles
    assert was_true



import logging
from pathlib import Path
from typing import Optional
from xml.etree import ElementTree

from cg.exc import FlowCellError
from typing_extensions import Literal

LOG = logging.getLogger(__name__)


class RunParameters:
    """Class to handle the run parameters from a sequencing run"""

    def __init__(self, run_parameters_path: Path):
        self.path: Path = run_parameters_path
        with open(run_parameters_path, "rt") as in_file:
            self.tree: ElementTree = ElementTree.parse(in_file)

    @property
    def index_length(self) -> int:
        index_one_length: int = self.index_read_one()
        index_two_length: int = self.index_read_two()
        if index_one_length != index_two_length:
            raise FlowCellError("Index lengths are not the same!")
        return index_one_length

    @property
    def control_software_version(self) -> str:
        node_name = ".ApplicationVersion"
        xml_node: Optional[ElementTree.Element] = self.tree.find(node_name)
        self.node_not_found(node=xml_node, name="control software version")
        return xml_node.text

    @property
    def reagent_kit_version(self) -> str:
        node_name = "./RfidsInfo/SbsConsumableVersion"
        xml_node: Optional[ElementTree.Element] = self.tree.find(node_name)
        if xml_node is None:
            LOG.warning("Could not determine reagent kit version")
            LOG.info("Set reagent kit version to 'unknown'")
            return "unknown"
        return xml_node.text

    @property
    def flowcell_type(self) -> Literal["novaseq", "hiseq"]:
        """Fetch the flowcell type from the run parameters"""
        # First try with the node name for hiseq
        node_name = "./Setup/ApplicationName"
        xml_node = self.tree.find(node_name)
        if xml_node is None:
            # Then try with node name for novaseq
            node_name = ".Application"
            xml_node = self.tree.find(node_name)
        self.node_not_found(node=xml_node, name="flowcell type")
        for flow_cell_name in ["novaseq", "hiseq"]:
            if flow_cell_name in xml_node.text.lower():
                return flow_cell_name
        message = f"Unknown flowcell type {xml_node.text}"
        LOG.warning(message)
        raise FlowCellError(message)

    @property
    def run_type(self) -> Literal["wgs", "fluffy"]:
        """Fetch what type of run the parameters is"""
        if self.index_length == 8:
            return "fluffy"
        return "wgs"

    @staticmethod
    def node_not_found(node: Optional[ElementTree.Element], name: str) -> None:
        """Raise exception if node if not found"""
        if node is None:
            message = f"Could not determine {name}"
            LOG.warning(message)
            raise FlowCellError(message)

    def get_node_integer_value(self, node_name: str, name: str) -> int:
        xml_node = self.tree.find(node_name)
        self.node_not_found(node=xml_node, name=name)
        return int(xml_node.text)

    def index_read_one(self) -> int:
        """Get the value for index read one"""
        node_name = "./IndexRead1NumberOfCycles"
        return self.get_node_integer_value(node_name=node_name, name="length of index one")

    def index_read_two(self) -> int:
        """Get the value for index read one"""
        node_name = "./IndexRead2NumberOfCycles"
        return self.get_node_integer_value(node_name=node_name, name="length of index two")

    def read_one_nr_cycles(self) -> int:
        """Get the nr of cycles for read one"""
        node_name = "./Read1NumberOfCycles"
        return self.get_node_integer_value(node_name=node_name, name="length of reads one")

    def read_two_nr_cycles(self) -> int:
        """Get the nr of cycles for read one"""
        node_name = "./Read2NumberOfCycles"
        return self.get_node_integer_value(node_name=node_name, name="length of reads two")

    def base_mask(self) -> str:
        """create the bcl2fastq basemask for novaseq flowcells

        Basemask is used in this comma format as an argument to bcl2fastq.
        When creating the unaligned path the commas are stripped
        """
        return (
            f"Y{self.read_one_nr_cycles()},"
            f"I{self.index_read_one()},"
            f"I{self.read_two_nr_cycles()},"
            f"Y{self.index_read_two()}"
        )

    def __str__(self):
        return f"RunParameters(path={self.path},flowcell_type={self.flowcell_type},run_type={self.run_type}"

    def __repr__(self):
        return (
            f"RunParameters(path={self.path},flowcell_type={self.flowcell_type},run_type={self.run_type},"
            f"reagent_kit_version={self.reagent_kit_version},control_software_version={self.control_software_version},"
            f"index_length={self.index_length})"
        )



from typing import Optional

from sqlalchemy import Column, orm, types
from sqlalchemy.orm.exc import NoResultFound

from .base import Model


class Flowcell(Model):
    flowcell_id = Column(types.Integer, primary_key=True)
    flowcellname = Column(types.String(255), nullable=False, unique=True)
    flowcell_pos = Column(types.Enum("A", "B"), nullable=False)
    hiseqtype = Column(types.String(255), nullable=False)
    time = Column(types.DateTime)

    demux = orm.relationship("Demux", cascade="all", backref=orm.backref("flowcell"))

    @staticmethod
    def exists(flowcell_name: str) -> Optional[int]:
        """Checks if the Flowcell entry already exists"""
        try:
            flowcell: Flowcell = Flowcell.query.filter_by(flowcellname=flowcell_name).one()
            return flowcell.flowcell_id
        except NoResultFound:
            return None



"""
    Module for interacting with crunchy to perform:
        1. Compressing: FASTQ to SPRING
        2. Decompressing: SPRING to FASTQ
    along with the helper methods.
"""

import datetime
import logging
from pathlib import Path
from typing import Dict, Optional

from cgmodels.crunchy.metadata import CrunchyFile, CrunchyMetadata

from cg.apps.crunchy import files
from cg.apps.slurm.slurm_api import SlurmAPI
from cg.constants import FASTQ_DELTA
from cg.models import CompressionData
from cg.models.slurm.sbatch import Sbatch

from cg.apps.crunchy.sbatch import (
    FASTQ_TO_SPRING_ERROR,
    SPRING_TO_FASTQ_ERROR,
    FASTQ_TO_SPRING_COMMANDS,
    SPRING_TO_FASTQ_COMMANDS,
)
from cg.constants.priority import SlurmQos

LOG = logging.getLogger(__name__)


class CrunchyAPI:
    """
    API for Crunchy.
    """

    def __init__(self, config: dict):
        self.conda_binary: Optional[str] = config["crunchy"]["conda_binary"] or None
        self.crunchy_env: str = config["crunchy"]["slurm"]["conda_env"]
        self.dry_run: bool = False
        self.reference_path: str = config["crunchy"]["cram_reference"]
        self.slurm_api: SlurmAPI = SlurmAPI()
        self.slurm_account: str = config["crunchy"]["slurm"]["account"]
        self.slurm_hours: int = config["crunchy"]["slurm"]["hours"]
        self.slurm_mail_user: str = config["crunchy"]["slurm"]["mail_user"]
        self.slurm_memory: int = config["crunchy"]["slurm"]["memory"]
        self.slurm_number_tasks: int = config["crunchy"]["slurm"]["number_tasks"]

    def set_dry_run(self, dry_run: bool) -> None:
        """Update dry run."""
        LOG.info("Updating compress api")
        LOG.info("Set dry run to %s", dry_run)
        self.dry_run = dry_run
        self.slurm_api.set_dry_run(dry_run=dry_run)

    # Methods to check compression status
    @staticmethod
    def is_compression_pending(compression_obj: CompressionData) -> bool:
        """Check if compression/decompression has started but not finished."""
        if compression_obj.pending_exists():
            LOG.info("Compression/decompression is pending for %s", compression_obj.run_name)
            return True
        LOG.info("Compression/decompression is not running")
        return False

    @staticmethod
    def is_fastq_compression_possible(compression_obj: CompressionData) -> bool:
        """Check if FASTQ compression is possible.

        There are three possible answers to this question:

         - Compression is running          -> Compression NOT possible
         - SPRING archive exists           -> Compression NOT possible
         - Data is external                -> Compression NOT possible
         - Not compressed and not running  -> Compression IS possible
        """
        if CrunchyAPI.is_compression_pending(compression_obj):
            return False

        if compression_obj.spring_exists():
            LOG.info("SPRING file found")
            return False

        if "external-data" in str(compression_obj.fastq_first):
            LOG.info("File is external data and should not be compressed")
            return False

        LOG.info("FASTQ compression is possible")

        return True

    @staticmethod
    def is_spring_decompression_possible(compression_obj: CompressionData) -> bool:
        """Check if SPRING decompression is possible.

        There are three possible answers to this question:

            - Compression/Decompression is running      -> Decompression is NOT possible
            - The FASTQ files are not compressed        -> Decompression is NOT possible
            - Compression has been performed            -> Decompression IS possible

        """
        if compression_obj.pending_exists():
            LOG.info("Compression/decompression is pending for %s", compression_obj.run_name)
            return False

        if not compression_obj.spring_exists():
            LOG.info("No SPRING file found")
            return False

        if compression_obj.pair_exists():
            LOG.info("FASTQ files already exists")
            return False

        LOG.info("Decompression is possible")

        return True

    @staticmethod
    def is_fastq_compression_done(compression: CompressionData) -> bool:
        """Check if FASTQ compression is finished.

        This is checked by controlling that the SPRING files that are produced after FASTQ
        compression exists.

        The following has to be fulfilled for FASTQ compression to be considered done:

            - A SPRING archive file exists
            - A SPRING archive metadata file exists
            - The SPRING archive has not been unpacked before FASTQ delta (21 days)

        Note:
        'updated_at' indicates at what date the SPRING archive was unarchived last.
        If the SPRING archive has never been unarchived 'updated_at' is None.

        """
        LOG.info("Check if FASTQ compression is finished")
        LOG.info(f"Check if SPRING file {compression.spring_path} exists")
        if not compression.spring_exists():
            LOG.info(
                f"No SPRING file for {compression.run_name}",
            )
            return False
        LOG.info("SPRING file found")

        LOG.info(f"Check if SPRING metadata file {compression.spring_metadata_path} exists")
        if not compression.metadata_exists():
            LOG.info("No metadata file found")
            return False
        LOG.info("SPRING metadata file found")

        # We want this to raise exception if file is malformed
        crunchy_metadata: CrunchyMetadata = files.get_crunchy_metadata(
            compression.spring_metadata_path
        )

        # Check if the SPRING archive has been unarchived
        updated_at: Optional[datetime.date] = files.get_file_updated_at(crunchy_metadata)
        if updated_at is None:
            LOG.info(f"FASTQ compression is done for {compression.run_name}")
            return True

        LOG.info(f"Files where unpacked {updated_at}")

        if not CrunchyAPI.check_if_update_spring(updated_at):
            return False

        LOG.info(f"FASTQ compression is done for {compression.run_name}")

        return True

    @staticmethod
    def is_spring_decompression_done(compression_obj: CompressionData) -> bool:
        """Check if SPRING decompression if finished.

        This means that all three files specified in SPRING metadata should exist.
        That is

            - First read in FASTQ pair should exist
            - Second read in FASTQ pair should exist
            - SPRING archive file should still exist
        """

        spring_metadata_path: Path = compression_obj.spring_metadata_path
        LOG.info(f"Check if SPRING metadata file {spring_metadata_path} exists")

        if not compression_obj.metadata_exists():
            LOG.info("No SPRING metadata file found")
            return False

        # We want this to exit hard if the metadata is malformed
        crunchy_metadata: CrunchyMetadata = files.get_crunchy_metadata(spring_metadata_path)

        for file_info in crunchy_metadata.files:
            if not Path(file_info.path).exists():
                LOG.info(f"File {file_info.path} does not exist")
                return False
            if not file_info.updated:
                LOG.info("Files have not been unarchived")
                return False

        LOG.info(f"SPRING decompression is done for run {compression_obj.run_name}")
        return True

    @staticmethod
    def create_pending_file(pending_path: Path, dry_run: bool) -> None:
        """Create a pending flag file."""
        LOG.info("Creating pending flag %s", pending_path)
        if dry_run:
            return
        pending_path.touch(exist_ok=False)

    # These are the compression/decompression methods
    def fastq_to_spring(self, compression_obj: CompressionData, sample_id: str = "") -> int:
        """Compress FASTQ files into SPRING by sending to sbatch SLURM."""
        CrunchyAPI.create_pending_file(
            pending_path=compression_obj.pending_path, dry_run=self.dry_run
        )
        log_dir: Path = files.get_log_dir(compression_obj.spring_path)
        # Generate the error function
        error_function = FASTQ_TO_SPRING_ERROR.format(
            spring_path=compression_obj.spring_path, pending_path=compression_obj.pending_path
        )
        # Generate the commands
        sbatch_parameters: Sbatch
        commands = FASTQ_TO_SPRING_COMMANDS.format(
            conda_run=f"{self.conda_binary} run --name {self.crunchy_env}",
            fastq_first=compression_obj.fastq_first,
            fastq_second=compression_obj.fastq_second,
            pending_path=compression_obj.pending_path,
            spring_path=compression_obj.spring_path,
            tmp_dir=files.get_tmp_dir(
                prefix="spring_",
                suffix="_compress",
                base=compression_obj.analysis_dir.as_posix(),
            ),
        )
        sbatch_parameters: Sbatch = Sbatch(
            account=self.slurm_account,
            commands=commands,
            email=self.slurm_mail_user,
            error=error_function,
            hours=self.slurm_hours,
            job_name="_".join([sample_id, compression_obj.run_name, "fastq_to_spring"]),
            log_dir=log_dir.as_posix(),
            memory=self.slurm_memory,
            number_tasks=self.slurm_number_tasks,
            quality_of_service=SlurmQos.MAINTENANCE,
        )
        sbatch_content: str = self.slurm_api.generate_sbatch_content(
            sbatch_parameters=sbatch_parameters
        )
        sbatch_path: Path = files.get_fastq_to_spring_sbatch_path(
            log_dir=log_dir, run_name=compression_obj.run_name
        )
        sbatch_number: int = self.slurm_api.submit_sbatch(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path
        )
        LOG.info("Fastq compression running as job %s", sbatch_number)
        return sbatch_number

    def spring_to_fastq(self, compression_obj: CompressionData, sample_id: str = "") -> int:
        """Decompress SPRING into FASTQ by submitting sbatch script to SLURM."""
        CrunchyAPI.create_pending_file(
            pending_path=compression_obj.pending_path, dry_run=self.dry_run
        )
        # Fetch the metadata information from a spring metadata file
        crunchy_metadata: CrunchyMetadata = files.get_crunchy_metadata(
            compression_obj.spring_metadata_path
        )
        files_info: Dict[str, CrunchyFile] = files.get_spring_archive_files(crunchy_metadata)
        log_dir = files.get_log_dir(compression_obj.spring_path)

        error_function = SPRING_TO_FASTQ_ERROR.format(
            fastq_first=compression_obj.fastq_first,
            fastq_second=compression_obj.fastq_second,
            pending_path=compression_obj.pending_path,
        )
        commands = SPRING_TO_FASTQ_COMMANDS.format(
            conda_run=f"{self.conda_binary} run --name {self.crunchy_env}",
            tmp_dir=files.get_tmp_dir(
                prefix="spring_",
                suffix="_decompress",
                base=compression_obj.analysis_dir.as_posix(),
            ),
            fastq_first=compression_obj.fastq_first,
            fastq_second=compression_obj.fastq_second,
            spring_path=compression_obj.spring_path,
            pending_path=compression_obj.pending_path,
            checksum_first=files_info["fastq_first"].checksum,
            checksum_second=files_info["fastq_second"].checksum,
        )
        sbatch_parameters: Sbatch = Sbatch(
            account=self.slurm_account,
            commands=commands,
            email=self.slurm_mail_user,
            error=error_function,
            hours=self.slurm_hours,
            job_name="_".join([sample_id, compression_obj.run_name, "spring_to_fastq"]),
            log_dir=log_dir.as_posix(),
            memory=self.slurm_memory,
            number_tasks=self.slurm_number_tasks,
            quality_of_service=SlurmQos.LOW,
        )
        sbatch_content: str = self.slurm_api.generate_sbatch_content(sbatch_parameters)
        sbatch_path = files.get_spring_to_fastq_sbatch_path(
            log_dir=log_dir, run_name=compression_obj.run_name
        )
        sbatch_number: int = self.slurm_api.submit_sbatch(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path
        )
        LOG.info("Spring decompression running as job %s", sbatch_number)
        return sbatch_number

    @staticmethod
    def check_if_update_spring(file_date: datetime.date) -> bool:
        """Check if date is older than FASTQ_DELTA."""
        delta = file_date + datetime.timedelta(days=FASTQ_DELTA)
        now = datetime.datetime.now()
        if delta > now.date():
            LOG.info("FASTQ files are not old enough")
            return False
        return True



"""Tests for helper functions in Cg Compress CLI."""

import logging
from pathlib import Path

from housekeeper.store.models import Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.cli.compress import helpers
from cg.cli.compress.helpers import set_memory_according_to_reads
from cg.constants.compression import MAX_READS_PER_GB, CRUNCHY_MIN_GB_PER_PROCESS


def test_set_memory_according_to_reads_when_no_reads(caplog, sample_id: str):
    """Test setting memory according to reads when no sample reads."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a sample id and no reads supplied

    # WHEN setting memory according to reads
    memory: int = set_memory_according_to_reads(sample_id=sample_id, sample_reads=0)

    # THEN we should log
    assert f"No reads recorded for sample: {sample_id}" in caplog.text

    # THEN None should be returned
    assert memory is None


def test_set_memory_according_to_reads_when_few_reads(sample_id: str):
    """Test setting memory according to reads when few reads."""
    # GIVEN a sample id and reads

    # WHEN setting memory according to reads
    memory: int = set_memory_according_to_reads(sample_id=sample_id, sample_reads=1)

    # THEN memory should be set to the minimum
    assert memory == CRUNCHY_MIN_GB_PER_PROCESS


def test_set_memory_according_to_reads_when_many_reads(sample_id: str):
    """Test setting memory according to reads when many reads."""
    # GIVEN a sample id and reads

    # WHEN setting memory according to reads
    memory: int = set_memory_according_to_reads(
        sample_id=sample_id, sample_reads=MAX_READS_PER_GB**10
    )

    # THEN memory should be limited to what is available on the node
    assert memory == 180


def test_set_memory_according_to_reads(sample_id: str):
    """Test setting memory according to reads."""
    # GIVEN a sample id and reads

    # WHEN setting memory according to reads
    memory: int = set_memory_according_to_reads(
        sample_id=sample_id, sample_reads=MAX_READS_PER_GB * 100
    )

    # THEN memory should be adjusted
    assert memory == 100


def test_get_true_dir_no_symlinks(project_dir: Path):
    """Test to get a true dir when there are no symlinks."""
    # GIVEN a directory with some files but no symlinked files
    a_file: Path = Path(project_dir, "hello.txt")
    a_file.touch()
    assert a_file.exists()

    # WHEN fetching the true dir for the files in fixture dir
    true_dir = helpers.get_true_dir(a_file.parent)

    # THEN assert that the true_dir is None since there where no symbolic links in the project_dir
    assert true_dir is None


def test_get_true_dir_when_symlinks(project_dir: Path):
    """Test to get a true dir when there are symlinks to another directory.

    The function should return the path to another directory when there are symlinks in one.
    """
    # GIVEN a directory with a file with some content
    content = "hello world"
    a_file: Path = Path(project_dir, "hello.txt")
    a_file.write_text(content)
    assert a_file.exists()

    # GIVEN a subdirectory with a link to the above file
    new_dir: Path = Path(project_dir, "new_dir/")
    new_dir.mkdir()
    a_link: Path = Path(new_dir, "link_to_hello")
    a_link.symlink_to(a_file)
    assert a_link.read_text() == content
    assert a_link.parent == new_dir
    assert a_link.is_symlink() is True

    # WHEN fetching the true dir for the symlinked dir
    true_dir = helpers.get_true_dir(a_link.parent)

    # THEN assert that the true_dir the same as the parent of the destination of the symlink
    assert true_dir == project_dir


def test_get_versions_no_bundle(housekeeper_api: HousekeeperAPI):
    """Test to get latest versions of bundles when no bundles exists."""
    # GIVEN a empty housekeeper_api

    # WHEN fetching versions
    versions: Iterator[Version] = helpers.get_versions(housekeeper_api)

    # THEN assert no versions was returned
    assert sum(1 for version in versions) == 0


def test_get_versions_one_bundle(housekeeper_api: HousekeeperAPI, spring_bundle: dict):
    """Test to get latest versions of bundles when no bundles exists."""
    # GIVEN a populated housekeeper_api
    housekeeper_api.add_bundle(spring_bundle)

    # WHEN fetching versions
    versions = helpers.get_versions(housekeeper_api)

    # THEN assert no versions was returned
    assert sum(1 for version in versions) == 1


def test_correct_spring_paths(
    housekeeper_api: HousekeeperAPI,
    spring_bundle_symlink_problem: dict,
    symlinked_fastqs: dict,
    new_dir: Path,
):
    """Test for correct_spring_paths."""
    # GIVEN a populated housekeeper_api
    housekeeper_api.add_bundle(spring_bundle_symlink_problem)
    # GIVEN that the spring files exists in the wrong location
    versions: Iterator[Version] = helpers.get_versions(housekeeper_api)
    version: Version = next(versions)
    for file_path in version.files:
        assert not Path(file_path.full_path).exists()

    # WHEN updating the spring paths
    helpers.correct_spring_paths(housekeeper_api)

    # THEN assert that the spring paths has been moved
    for file_path in version.files:
        assert Path(file_path.full_path).exists()



"""Tests enabled commands for the generate command"""

from cg.constants import EXIT_SUCCESS

from cg.cli.generate.base import generate


def test_generate_no_options(base_context, cli_runner):
    """Tests generate command with no options"""

    # GIVEN a base context

    # WHEN dry running
    result = cli_runner.invoke(generate, obj=base_context)

    # THEN the output should list all the supported pipelines
    assert result.exit_code == EXIT_SUCCESS
    assert "delivery-report" in result.output
    assert "available-delivery-reports" in result.output



"""cg module for cleaning databases and files."""
import logging
from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Optional
from cg.utils.dispatcher import Dispatcher

import click
from alchy import Query
from cgmodels.cg.constants import Pipeline
from housekeeper.store.models import File, Version
from tabulate import tabulate

from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.scout.scout_export import ScoutExportCase
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.cli.workflow.commands import (
    balsamic_past_run_dirs,
    balsamic_qc_past_run_dirs,
    balsamic_umi_past_run_dirs,
    balsamic_pon_past_run_dirs,
    fluffy_past_run_dirs,
    mip_dna_past_run_dirs,
    mip_rna_past_run_dirs,
    mutant_past_run_dirs,
    rnafusion_past_run_dirs,
    rsync_past_run_dirs,
    microsalt_past_run_dirs,
)
from cg.constants import FlowCellStatus
from cg.constants.constants import DRY_RUN, SKIP_CONFIRMATION
from cg.constants.sequencing import Sequencers
from cg.constants.housekeeper_tags import SequencingFileTag, ALIGNMENT_FILE_TAGS, ScoutTag
from cg.models.demultiplex.flow_cell import FlowCell as DemultiplexFlowCell
from cg.utils.date import get_timedelta_from_date, get_date_days_ago
from cg.exc import FlowCellError, HousekeeperBundleVersionMissingError
from cg.meta.clean.api import CleanAPI
from cg.meta.clean.demultiplexed_flow_cells import DemultiplexedRunsFlowCell
from cg.meta.clean.flow_cell_run_directories import RunDirFlowCell
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Sample, Flowcell, Analysis

CHECK_COLOR = {True: "green", False: "red"}
LOG = logging.getLogger(__name__)
FLOW_CELL_OUTPUT_HEADERS = [
    "Flow cell run name",
    "Flow cell id",
    "Correct name?",
    "Exists in statusdb?",
    "Fastq files in HK?",
    "Spring files in HK?",
    "Files on disk?",
    "Check passed?",
]


@click.group()
def clean():
    """Clean up processes."""
    return


for sub_cmd in [
    balsamic_past_run_dirs,
    balsamic_qc_past_run_dirs,
    balsamic_umi_past_run_dirs,
    balsamic_pon_past_run_dirs,
    fluffy_past_run_dirs,
    mip_dna_past_run_dirs,
    mip_rna_past_run_dirs,
    mutant_past_run_dirs,
    rnafusion_past_run_dirs,
    rsync_past_run_dirs,
    microsalt_past_run_dirs,
]:
    clean.add_command(sub_cmd)


@clean.command("hk-alignment-files")
@click.argument("bundle")
@DRY_RUN
@SKIP_CONFIRMATION
@click.pass_obj
def hk_alignment_files(
    context: CGConfig, bundle: str, yes: bool = False, dry_run: bool = False
) -> None:
    """Clean up alignment files in Housekeeper bundle."""
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    for tag in ALIGNMENT_FILE_TAGS:
        tag_files = set(housekeeper_api.get_files(bundle=bundle, tags=[tag]))

        if not tag_files:
            LOG.warning(
                f"Could not find any files ready for cleaning for bundle {bundle} and tag {tag}"
            )

        for hk_file in tag_files:
            if not (yes or click.confirm(_get_confirm_question(bundle, hk_file))):
                continue

            file_path: Path = Path(hk_file.full_path)
            if hk_file.is_included and file_path.exists():
                LOG.info(f"Unlinking {file_path}")
                if not dry_run:
                    file_path.unlink()

            LOG.info(f"Deleting {file_path} from database")
            if not dry_run:
                hk_file.delete()
                housekeeper_api.commit()


@clean.command("scout-finished-cases")
@click.option(
    "--days-old",
    type=int,
    default=300,
    help="Clean alignment files with analysis dates older then given number of days",
)
@SKIP_CONFIRMATION
@DRY_RUN
@click.pass_context
def scout_finished_cases(
    context: click.Context, days_old: int, yes: bool = False, dry_run: bool = False
) -> None:
    """Clean up of solved and archived Scout cases."""
    scout_api: ScoutAPI = context.obj.scout_api
    bundles: List[str] = []
    for status in [ScoutTag.ARCHIVED.value, ScoutTag.SOLVED.value]:
        cases: List[ScoutExportCase] = scout_api.get_cases(status=status, reruns=False)
        cases_added: int = 0
        for case in cases:
            analysis_time_delta: timedelta = get_timedelta_from_date(date=case.analysis_date)
            if analysis_time_delta.days > days_old:
                bundles.append(case.id)
                cases_added += 1
        LOG.info(f"{cases_added} cases marked for alignment files removal")

    for bundle in bundles:
        context.invoke(hk_alignment_files, bundle=bundle, yes=yes, dry_run=dry_run)


@clean.command("hk-case-bundle-files")
@click.option(
    "--days-old",
    type=int,
    default=365,
    help="Clean all files with analysis dates older then given number of days",
)
@DRY_RUN
@click.pass_context
def hk_case_bundle_files(context: CGConfig, days_old: int, dry_run: bool = False) -> None:
    """Clean up all non-protected files for all pipelines."""
    housekeeper_api: HousekeeperAPI = context.obj.housekeeper_api
    clean_api: CleanAPI = CleanAPI(status_db=context.obj.status_db, housekeeper_api=housekeeper_api)

    size_cleaned: int = 0
    version_file: File
    for version_file in clean_api.get_unprotected_existing_bundle_files(
        before=get_date_days_ago(days_ago=days_old)
    ):
        file_path: Path = Path(version_file.full_path)
        file_size: int = file_path.stat().st_size
        size_cleaned += file_size
        if dry_run:
            LOG.info(f"Dry run: {dry_run}. Keeping file {file_path}")
            continue

        file_path.unlink()
        housekeeper_api.delete_file(file_id=version_file.id)
        housekeeper_api.commit()
        LOG.info(f"Removed file {file_path}. Dry run: {dry_run}")

    LOG.info(f"Process freed {round(size_cleaned * 0.0000000001, 2)} GB. Dry run: {dry_run}")


@clean.command("hk-bundle-files")
@click.option("-c", "--case-id", type=str, required=False)
@click.option("-p", "--pipeline", type=Pipeline, required=False)
@click.option("-t", "--tags", multiple=True, required=True)
@click.option("-o", "--days-old", type=int, default=30)
@DRY_RUN
@click.pass_obj
def hk_bundle_files(
    context: CGConfig,
    case_id: Optional[str],
    tags: list,
    days_old: Optional[int],
    pipeline: Optional[Pipeline],
    dry_run: bool,
):
    """Remove files found in Housekeeper bundles."""

    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    status_db: Store = context.status_db

    date_threshold: datetime = get_date_days_ago(days_ago=days_old)

    function_dispatcher: Dispatcher = Dispatcher(
        functions=[
            status_db.get_analyses_started_at_before,
            status_db.get_analyses_for_case_and_pipeline_started_at_before,
            status_db.get_analyses_for_pipeline_started_at_before,
            status_db.get_analyses_for_case_started_at_before,
        ],
        input_dict={
            "case_internal_id": case_id,
            "pipeline": pipeline,
            "started_at_before": date_threshold,
        },
    )
    analyses: List[Analysis] = function_dispatcher()

    size_cleaned: int = 0
    for analysis in analyses:
        LOG.info(f"Cleaning analysis {analysis}")
        bundle_name: str = analysis.family.internal_id
        hk_bundle_version: Optional[Version] = housekeeper_api.version(
            bundle=bundle_name, date=analysis.started_at
        )
        if not hk_bundle_version:
            LOG.warning(
                f"Version not found for "
                f"bundle:{bundle_name}; "
                f"pipeline: {analysis.pipeline}; "
                f"date {analysis.started_at}"
            )
            continue

        LOG.info(
            f"Version found for "
            f"bundle:{bundle_name}; "
            f"pipeline: {analysis.pipeline}; "
            f"date {analysis.started_at}"
        )
        version_files: List[File] = housekeeper_api.get_files(
            bundle=analysis.family.internal_id, tags=tags, version=hk_bundle_version.id
        ).all()
        for version_file in version_files:
            file_path: Path = Path(version_file.full_path)
            if not file_path.exists():
                LOG.info(f"File {file_path} not on disk.")
                continue
            LOG.info(f"File {file_path} found on disk.")
            file_size = file_path.stat().st_size
            size_cleaned += file_size
            if dry_run:
                continue

            file_path.unlink()
            housekeeper_api.delete_file(version_file.id)
            housekeeper_api.commit()
            LOG.info(f"Removed file {file_path}. Dry run: {dry_run}")

    LOG.info(f"Process freed {round(size_cleaned * 0.0000000001, 2)}GB. Dry run: {dry_run}")


@clean.command("invalid-flow-cell-dirs")
@click.option("-f", "--failed-only", is_flag=True, help="Shows failed flow cells only")
@DRY_RUN
@click.pass_obj
def remove_invalid_flow_cell_directories(context: CGConfig, failed_only: bool, dry_run: bool):
    """Removes invalid flow cell directories from demultiplexed-runs"""
    status_db: Store = context.status_db
    demux_api: DemultiplexingAPI = context.demultiplex_api
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    trailblazer_api: TrailblazerAPI = context.trailblazer_api
    sample_sheets_dir: str = context.clean.flow_cells.sample_sheets_dir_name
    checked_flow_cells: List[DemultiplexedRunsFlowCell] = []
    search: str = f"%{demux_api.out_dir}%"
    fastq_files_in_housekeeper: Query = housekeeper_api.files(
        tags=[SequencingFileTag.FASTQ]
    ).filter(File.path.like(search))
    spring_files_in_housekeeper: Query = housekeeper_api.files(
        tags=[SequencingFileTag.SPRING]
    ).filter(File.path.like(search))
    for flow_cell_dir in demux_api.out_dir.iterdir():
        flow_cell: DemultiplexedRunsFlowCell = DemultiplexedRunsFlowCell(
            flow_cell_path=flow_cell_dir,
            status_db=status_db,
            housekeeper_api=housekeeper_api,
            trailblazer_api=trailblazer_api,
            sample_sheets_dir=sample_sheets_dir,
            fastq_files=fastq_files_in_housekeeper,
            spring_files=spring_files_in_housekeeper,
        )
        if not flow_cell.is_demultiplexing_ongoing_or_started_and_not_completed:
            LOG.info(f"Found flow cell ready to be checked: {flow_cell.path}!")
            checked_flow_cells.append(flow_cell)
            if not flow_cell.passed_check:
                LOG.warning(f"Invalid flow cell directory found: {flow_cell.path}")
                if dry_run:
                    continue
                LOG.warning(f"Removing {flow_cell.path}!")
                flow_cell.remove_failed_flow_cell()
        else:
            LOG.warning("Skipping check!")

    failed_flow_cells: List[DemultiplexedRunsFlowCell] = [
        flow_cell for flow_cell in checked_flow_cells if not flow_cell.passed_check
    ]
    flow_cells_to_present: List[DemultiplexedRunsFlowCell] = (
        failed_flow_cells if failed_only else checked_flow_cells
    )
    tabulate_row = [
        [
            flow_cell.run_name,
            flow_cell.id,
            flow_cell.is_correctly_named,
            flow_cell.exists_in_statusdb,
            flow_cell.fastq_files_exist_in_housekeeper,
            flow_cell.spring_files_exist_in_housekeeper,
            flow_cell.files_exist_on_disk,
            click.style(str(flow_cell.passed_check), fg=CHECK_COLOR[flow_cell.passed_check]),
        ]
        for flow_cell in flow_cells_to_present
    ]

    click.echo(
        tabulate(
            tabulate_row,
            headers=FLOW_CELL_OUTPUT_HEADERS,
            tablefmt="fancy_grid",
            missingval="N/A",
        ),
    )


@clean.command("fix-flow-cell-status")
@DRY_RUN
@click.pass_obj
def fix_flow_cell_status(context: CGConfig, dry_run: bool):
    """Set correct flow cell statuses in Statusdb."""
    status_db: Store = context.status_db
    housekeeper_api: HousekeeperAPI = context.housekeeper_api

    flow_cells_in_statusdb: List[Flowcell] = list(
        status_db.get_flow_cells_by_statuses(
            flow_cell_statuses=[FlowCellStatus.ON_DISK, FlowCellStatus.REMOVED]
        )
    )

    LOG.info(
        f"Number of flow cells with status {FlowCellStatus.ON_DISK.value} or {FlowCellStatus.REMOVED} in Statusdb: {len(flow_cells_in_statusdb)}"
    )

    for flow_cell in flow_cells_in_statusdb:
        sample_bundle_names: List[str] = [sample.internal_id for sample in flow_cell.samples]
        are_sequencing_files_in_hk: bool = False
        are_sequencing_files_on_disk: bool = False
        try:
            are_sequencing_files_in_hk: bool = housekeeper_api.is_fastq_or_spring_in_all_bundles(
                bundle_names=sample_bundle_names
            )
            are_sequencing_files_on_disk: bool = (
                housekeeper_api.is_fastq_or_spring_on_disk_in_all_bundles(
                    bundle_names=sample_bundle_names
                )
            )
        except HousekeeperBundleVersionMissingError:
            LOG.warning(
                f"Cannot find sample bundle in Housekeeper for sample on flow cell: {flow_cell.name}"
            )
        new_status: str = (
            FlowCellStatus.ON_DISK
            if are_sequencing_files_in_hk and are_sequencing_files_on_disk
            else FlowCellStatus.REMOVED
        )

        if flow_cell.status != new_status:
            LOG.info(
                f"Setting status of flow cell {flow_cell.name} from: {flow_cell.status} to {new_status}"
            )
            if dry_run:
                continue
            flow_cell.status: str = new_status
            status_db.commit()


@clean.command("remove-old-flow-cell-run-dirs")
@click.option(
    "-s",
    "--sequencer",
    type=click.Choice([Sequencers.HISEQX, Sequencers.HISEQGA, Sequencers.NOVASEQ, Sequencers.ALL]),
    default="all",
    help="Specify the sequencer. Default is to remove flow cells for all sequencers",
)
@click.option(
    "-o",
    "--days-old",
    type=int,
    default=21,
    help="Specify the age in days of the flow cells to be removed.",
)
@DRY_RUN
@click.pass_obj
def remove_old_flow_cell_run_dirs(context: CGConfig, sequencer: str, days_old: int, dry_run: bool):
    """Removes flow cells from flow cell run dir based on the sequencing date and
    the sequencer type, if specified."""
    status_db: Store = context.status_db
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    if sequencer == Sequencers.ALL:
        LOG.info("Checking flow cells for all sequencers!")
        for sequencer, run_directory in context.clean.flow_cells.flow_cell_run_dirs:
            LOG.info(f"Checking directory {run_directory} of sequencer {sequencer}:")
            flow_cell_dirs: List[Path] = [
                flow_cell_dir
                for flow_cell_dir in Path(run_directory).iterdir()
                if flow_cell_dir.is_dir()
            ]
            for flow_cell_dir in flow_cell_dirs:
                clean_run_directories(
                    days_old=days_old,
                    dry_run=dry_run,
                    housekeeper_api=housekeeper_api,
                    run_directory=flow_cell_dir,
                    status_db=status_db,
                )

    else:
        run_directory: str = dict(context.clean.flow_cells.flow_cell_run_dirs).get(sequencer)
        LOG.info(f"Checking directory {run_directory} of sequencer {sequencer}:")
        flow_cell_dirs: List[Path] = [
            flow_cell_dir
            for flow_cell_dir in Path(run_directory).iterdir()
            if flow_cell_dir.is_dir()
        ]
        for flow_cell_dir in flow_cell_dirs:
            clean_run_directories(
                days_old=days_old,
                dry_run=dry_run,
                housekeeper_api=housekeeper_api,
                run_directory=flow_cell_dir,
                status_db=status_db,
            )


@clean.command("remove-old-demutliplexed-run-dirs")
@click.option(
    "-o",
    "--days-old",
    type=int,
    default=21,
    help="Specify the age in days of the flow cells to be removed.",
)
@DRY_RUN
@click.pass_obj
def remove_old_demutliplexed_run_dirs(context: CGConfig, days_old: int, dry_run: bool):
    """Removes flow cells from demultiplexed run directory."""
    status_db: Store = context.status_db
    demux_api: DemultiplexingAPI = context.demultiplex_api
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    trailblazer_api: TrailblazerAPI = context.trailblazer_api
    for flow_cell_dir in demux_api.get_all_demultiplexed_flow_cell_dirs():
        try:
            flow_cell: DemultiplexFlowCell = DemultiplexFlowCell(flow_cell_path=flow_cell_dir)
        except FlowCellError:
            continue
        samples: List[Sample] = status_db.get_samples_from_flow_cell(flow_cell_id=flow_cell.id)
        are_sequencing_files_in_hk: bool = housekeeper_api.is_fastq_or_spring_in_all_bundles(
            bundle_names=[sample.internal_id for sample in samples]
        )
        demux_runs_flow_cell: DemultiplexedRunsFlowCell = DemultiplexedRunsFlowCell(
            flow_cell_path=flow_cell_dir,
            status_db=status_db,
            housekeeper_api=housekeeper_api,
            trailblazer_api=trailblazer_api,
            sample_sheets_dir=context.clean.flow_cells.sample_sheets_dir_name,
        )
        if (
            not demux_runs_flow_cell.is_demultiplexing_ongoing_or_started_and_not_completed
            and are_sequencing_files_in_hk
        ):
            LOG.info(f"Found flow cell ready to be removed: {demux_runs_flow_cell.path}!")
            if dry_run:
                continue
            LOG.warning(f"Removing {demux_runs_flow_cell.path}!")
            clean_run_directories(
                days_old=days_old,
                dry_run=dry_run,
                housekeeper_api=housekeeper_api,
                run_directory=demux_runs_flow_cell.path,
                status_db=status_db,
            )


def clean_run_directories(
    days_old: int,
    dry_run: bool,
    housekeeper_api: HousekeeperAPI,
    run_directory: Path,
    status_db: Store,
):
    """Cleans up all flow cell directories in the specified run directory."""
    LOG.info(f"Checking flow cell {run_directory.name}")
    run_dir_flow_cell: RunDirFlowCell = RunDirFlowCell(
        flow_cell_dir=run_directory, status_db=status_db, housekeeper_api=housekeeper_api
    )
    if run_dir_flow_cell.exists_in_statusdb and run_dir_flow_cell.is_retrieved_from_pdc:
        LOG.info(
            f"Skipping removal of flow cell {run_directory}, PDC retrieval status is '{run_dir_flow_cell.flow_cell_status}'!"
        )
        return
    if run_dir_flow_cell.age < days_old:
        LOG.info(
            f"Flow cell {run_directory} is {run_dir_flow_cell.age} days old and will NOT be removed."
        )
        return
    LOG.info(f"Flow cell {run_directory} is {run_dir_flow_cell.age} days old and will be removed.")
    if dry_run:
        return
    LOG.info(f"Removing flow cell run directory {run_dir_flow_cell.flow_cell_dir}.")
    run_dir_flow_cell.archive_sample_sheet()
    run_dir_flow_cell.remove_run_directory()


def _get_confirm_question(bundle, file_obj) -> str:
    """Return confirmation question."""
    return (
        f"{bundle}: remove file from file system and database: {file_obj.full_path}"
        if file_obj.is_included
        else f"{bundle}: remove file from database: {file_obj.full_path}"
    )



"""Tests the cli for uploading delivery reports"""

from cg.cli.upload.delivery_report import upload_delivery_report_to_scout
from cg.constants import EXIT_FAIL, EXIT_SUCCESS
from cg.meta.report.mip_dna import MipDNAReportAPI


def test_delivery_report_to_scout_no_params(upload_context, cli_runner):
    """Tests the upload to Scout without specifying the case"""

    # GIVEN a MIP-DNA report api
    assert isinstance(upload_context.meta_apis.get("report_api"), MipDNAReportAPI)

    # WHEN invoking the delivery report upload without parameters
    result = cli_runner.invoke(
        upload_delivery_report_to_scout,
        obj=upload_context,
    )

    # THEN the command should fail due to a missing case ID
    assert "There are no valid cases to perform delivery report actions" in result.output
    assert result.exit_code == EXIT_FAIL


def test_delivery_report_to_scout(upload_context, cli_runner, upload_report_hk_api, case_id):
    """Tests the upload to Scout of a MIP DNA delivery report"""

    # GIVEN a Housekeeper context with a delivery report file that is ready for upload
    upload_context.housekeeper_api_ = upload_report_hk_api

    # WHEN uploading the delivery report
    result = cli_runner.invoke(
        upload_delivery_report_to_scout,
        [case_id],
        obj=upload_context,
    )

    # THEN check that the command exits with success and that the mock file has been uploaded to Scout
    assert result.exit_code == EXIT_SUCCESS



"""Tests for the meta BackupAPI"""

import logging
import subprocess

import mock
import pytest
from mock import call

from cg.constants.sequencing import Sequencers
from tests.mocks.hk_mock import MockFile

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import FileExtensions, FlowCellStatus
from cg.exc import ChecksumFailedError
from cg.meta.backup.backup import BackupAPI, SpringBackupAPI
from cg.meta.backup.pdc import PdcAPI
from cg.meta.encryption.encryption import SpringEncryptionAPI


@mock.patch("cg.store.Store")
def test_maximum_processing_queue_full(mock_store):
    """Tests check_processing method of the backup API"""
    # GIVEN a flow cell needs to be retrieved from PDC
    backup_api = BackupAPI(
        encryption_api=mock.Mock(),
        encrypt_dir=mock.Mock(),
        status=mock_store,
        tar_api=mock.Mock(),
        pdc_api=mock.Mock(),
        root_dir=mock.Mock(),
    )

    # WHEN there's already a flow cell being retrieved from PDC
    mock_store.get_flow_cells_by_statuses(
        flow_cell_statuses=[FlowCellStatus.PROCESSING]
    ).count.return_value = 1

    # THEN this method should return False
    assert backup_api.check_processing() is False


@mock.patch("cg.store")
def test_maximum_processing_queue_not_full(mock_store):
    """Tests check_processing method of the backup API"""
    # GIVEN a flow cell needs to be retrieved from PDC
    backup_api = BackupAPI(
        encryption_api=mock.Mock(),
        encrypt_dir=mock.Mock(),
        status=mock_store,
        tar_api=mock.Mock(),
        pdc_api=mock.Mock(),
        root_dir=mock.Mock(),
    )
    # WHEN there are no flow cells being retrieved from PDC
    mock_store.get_flow_cells_by_statuses(
        flow_cell_statuses=[FlowCellStatus.PROCESSING]
    ).count.return_value = 0

    # THEN this method should return True
    assert backup_api.check_processing() is True


@mock.patch("cg.store.models.Flowcell")
@mock.patch("cg.store")
def test_get_first_flow_cell_next_requested(mock_store, mock_flow_cell):
    """Tests get_first_flow_cell method of the backup API when requesting next flow cell"""
    # GIVEN status-db needs to be checked for flow cells to be retrieved from PDC
    backup_api = BackupAPI(
        encryption_api=mock.Mock(),
        encrypt_dir=mock.Mock(),
        status=mock_store,
        tar_api=mock.Mock(),
        pdc_api=mock.Mock(),
        root_dir=mock.Mock(),
    )

    # WHEN a flow cell is requested to be retrieved from PDC
    mock_store.get_flow_cells_by_statuses(
        flow_cell_statuses=[FlowCellStatus.REQUESTED]
    ).first.return_value = mock_flow_cell

    popped_flow_cell = backup_api.get_first_flow_cell()

    # THEN a flow cell is returned
    assert popped_flow_cell is not None


@mock.patch("cg.store")
def test_get_first_flow_cell_no_flow_cell_requested(mock_store):
    """Tests get_first_flow_cell method of the backup API when no flow cell requested"""
    # GIVEN status-db needs to be checked for flow cells to be retrieved from PDC
    backup_api = BackupAPI(
        encryption_api=mock.Mock(),
        encrypt_dir=mock.Mock(),
        status=mock_store,
        tar_api=mock.Mock(),
        pdc_api=mock.Mock(),
        root_dir=mock.Mock(),
    )

    # WHEN there are no flow cells requested to be retrieved from PDC
    mock_store.get_flow_cells_by_statuses(
        flow_cell_statuses=[FlowCellStatus.REQUESTED]
    ).first.return_value = None

    popped_flow_cell = backup_api.get_first_flow_cell()

    # THEN no flow cell is returned
    assert popped_flow_cell is None


@mock.patch("cg.meta.backup.backup.BackupAPI.check_processing")
@mock.patch("cg.store.models.Flowcell")
def test_fetch_flow_cell_processing_queue_full(mock_flow_cell, mock_check_processing, caplog):
    """Tests the fetch_flow_cell method of the backup API when processing queue is full"""

    caplog.set_level(logging.INFO)

    # GIVEN we check if a flow cell needs to be retrieved from PDC
    backup_api = BackupAPI(
        encryption_api=mock.Mock(),
        encrypt_dir=mock.Mock(),
        status=mock.Mock(),
        tar_api=mock.Mock(),
        pdc_api=mock.Mock(),
        root_dir=mock.Mock(),
    )

    # WHEN the processing queue is full
    backup_api.check_processing.return_value = False
    result = backup_api.fetch_flow_cell(mock_flow_cell)

    # THEN no flow cell will be fetched and a log message indicates that the processing queue is
    # full
    assert result is None
    assert "Processing queue is full" in caplog.text


@mock.patch("cg.meta.backup.backup.BackupAPI.get_first_flow_cell")
@mock.patch("cg.meta.backup.backup.BackupAPI.check_processing")
@mock.patch("cg.store")
def test_fetch_flow_cell_no_flow_cells_requested(
    mock_store,
    mock_check_processing,
    mock_get_first_flow_cell,
    caplog,
):
    """Tests the fetch_flow_cell method of the backup API when no flow cell requested"""

    caplog.set_level(logging.INFO)

    # GIVEN we check if a flow cell needs to be retrieved from PDC
    backup_api = BackupAPI(
        encryption_api=mock.Mock(),
        encrypt_dir=mock.Mock(),
        status=mock_store,
        tar_api=mock.Mock(),
        pdc_api=mock.Mock(),
        root_dir=mock.Mock(),
    )

    # WHEN no flow cells are requested
    backup_api.get_first_flow_cell.return_value = None
    backup_api.check_processing.return_value = True

    # AND no flow cell has been specified
    mock_flow_cell = None

    result = backup_api.fetch_flow_cell(mock_flow_cell)

    # THEN no flow cell will be fetched and a log message indicates that no flow cells have been
    # requested
    assert result is None
    assert "No flow cells requested" in caplog.text


@mock.patch("cg.meta.backup.backup.BackupAPI.unlink_files")
@mock.patch("cg.meta.backup.backup.BackupAPI.create_rta_complete")
@mock.patch("cg.meta.backup.backup.BackupAPI.get_archived_flow_cell_path")
@mock.patch("cg.meta.backup.backup.BackupAPI.get_archived_encryption_key_path")
@mock.patch("cg.meta.backup.backup.BackupAPI.check_processing")
@mock.patch("cg.meta.backup.backup.BackupAPI.get_first_flow_cell")
@mock.patch("cg.meta.tar.tar.TarAPI")
@mock.patch("cg.store.models.Flowcell")
@mock.patch("cg.store")
def test_fetch_flow_cell_retrieve_next_flow_cell(
    mock_store,
    mock_flow_cell,
    mock_tar,
    mock_get_first_flow_cell,
    mock_get_archived_encryption_key_path,
    mock_get_archived_flow_cell_path,
    mock_create_rta_complete,
    mock_check_processing,
    mock_unlink_files,
    archived_key,
    archived_flow_cell,
    cg_context,
    caplog,
):
    """Tests the fetch_flow_cell method of the backup API when retrieving next flow cell"""

    caplog.set_level(logging.INFO)

    # GIVEN we check if a flow cell needs to be retrieved from PDC
    backup_api = BackupAPI(
        encryption_api=mock.Mock(),
        encrypt_dir=cg_context.backup.encrypt_dir.dict(),
        status=mock_store,
        tar_api=mock_tar,
        pdc_api=mock.Mock(),
        root_dir=cg_context.backup.root.dict(),
    )

    # WHEN no flow cell is specified, but a flow cell in status-db has the status "requested"
    mock_flow_cell.status = FlowCellStatus.REQUESTED
    mock_flow_cell.sequencer_type = Sequencers.NOVASEQ
    backup_api.get_first_flow_cell.return_value = mock_flow_cell
    backup_api.check_processing.return_value = True
    backup_api.get_archived_encryption_key_path.return_value = archived_key
    backup_api.get_archived_flow_cell_path.return_value = archived_flow_cell
    backup_api.tar_api.run_tar_command.return_value = None
    result = backup_api.fetch_flow_cell(flow_cell=None)

    # THEN the process to retrieve the flow cell from PDC is started
    assert "retrieving from PDC" in caplog.text

    # AND when done the status of that flow cell is set to "retrieved"
    assert (
        f"Status for flow cell {mock_get_first_flow_cell.return_value.name} set to "
        f"{FlowCellStatus.RETRIEVED}" in caplog.text
    )
    assert mock_flow_cell.status == "retrieved"

    # AND status-db is updated with the new status
    assert mock_store.commit.called

    # AND the elapsed time of the retrieval process is returned
    assert result > 0


@mock.patch("cg.meta.backup.backup.BackupAPI.unlink_files")
@mock.patch("cg.meta.backup.backup.BackupAPI.create_rta_complete")
@mock.patch("cg.meta.backup.backup.BackupAPI.get_archived_flow_cell_path")
@mock.patch("cg.meta.backup.backup.BackupAPI.get_archived_encryption_key_path")
@mock.patch("cg.meta.backup.backup.BackupAPI.check_processing")
@mock.patch("cg.meta.backup.backup.BackupAPI.get_first_flow_cell")
@mock.patch("cg.meta.tar.tar.TarAPI")
@mock.patch("cg.store.models.Flowcell")
@mock.patch("cg.store")
def test_fetch_flow_cell_retrieve_specified_flow_cell(
    mock_store,
    mock_flow_cell,
    mock_tar,
    mock_get_first_flow_cell,
    mock_check_processing,
    mock_get_archived_key,
    mock_get_archived_flow_cell,
    mock_create_rta_complete,
    mock_unlink_files,
    archived_key,
    archived_flow_cell,
    cg_context,
    caplog,
):
    """Tests the fetch_flow_cell method of the backup API when given a flow cell"""

    caplog.set_level(logging.INFO)

    # GIVEN we want to retrieve a specific flow cell from PDC
    backup_api = BackupAPI(
        encryption_api=mock.Mock(),
        encrypt_dir=cg_context.backup.encrypt_dir.dict(),
        status=mock_store,
        tar_api=mock_tar,
        pdc_api=mock.Mock(),
        root_dir=cg_context.backup.root.dict(),
    )
    mock_flow_cell.status = FlowCellStatus.REQUESTED
    mock_flow_cell.sequencer_type = Sequencers.NOVASEQ
    backup_api.check_processing.return_value = True
    backup_api.get_archived_encryption_key_path.return_value = archived_key
    backup_api.get_archived_flow_cell_path.return_value = archived_flow_cell
    backup_api.tar_api.run_tar_command.return_value = None
    result = backup_api.fetch_flow_cell(flow_cell=mock_flow_cell)

    # THEN no flow cell is taken form statusdb
    mock_get_first_flow_cell.assert_not_called()

    # THEN the process to retrieve the flow cell from PDC is started
    assert "retrieving from PDC" in caplog.text

    # AND when done the status of that flow cell is set to "retrieved"
    assert (
        f"Status for flow cell {mock_flow_cell.name} set to {FlowCellStatus.RETRIEVED}"
        in caplog.text
    )
    assert mock_flow_cell.status == "retrieved"

    # AND status-db is updated with the new status
    assert mock_store.commit.called

    # AND the elapsed time of the retrieval process is returned
    assert result > 0


@mock.patch("cg.meta.backup.backup.BackupAPI.unlink_files")
@mock.patch("cg.meta.backup.backup.BackupAPI.create_rta_complete")
@mock.patch("cg.meta.backup.backup.BackupAPI.get_archived_flow_cell_path")
@mock.patch("cg.meta.backup.backup.BackupAPI.get_archived_encryption_key_path")
@mock.patch("cg.meta.backup.backup.BackupAPI.check_processing")
@mock.patch("cg.meta.backup.backup.BackupAPI.get_first_flow_cell")
@mock.patch("cg.meta.tar.tar.TarAPI")
@mock.patch("cg.store.models.Flowcell")
@mock.patch("cg.meta.backup.pdc.PdcAPI")
@mock.patch("cg.store")
def test_fetch_flow_cell_pdc_retrieval_failed(
    mock_store,
    mock_pdc,
    mock_flow_cell,
    mock_tar,
    mock_get_first_flow_cell,
    mock_check_processing,
    mock_get_archived_key,
    mock_get_archived_flow_cell,
    archived_key,
    archived_flow_cell,
    cg_context,
    caplog,
):
    """Tests the fetch_flow_cell method of the backup API when PDC retrieval failed"""

    caplog.set_level(logging.INFO)

    # GIVEN we are going to retrieve a flow cell from PDC
    backup_api = BackupAPI(
        encryption_api=mock.Mock(),
        encrypt_dir=cg_context.backup.encrypt_dir.dict(),
        status=mock_store,
        tar_api=mock_tar,
        pdc_api=mock_pdc,
        root_dir=cg_context.backup.root.dict(),
    )
    mock_flow_cell.status = FlowCellStatus.REQUESTED
    mock_flow_cell.sequencer_type = Sequencers.NOVASEQ
    backup_api.check_processing.return_value = True
    backup_api.get_archived_encryption_key_path.return_value = archived_key
    backup_api.get_archived_flow_cell_path.return_value = archived_flow_cell
    backup_api.tar_api.run_tar_command.return_value = None

    # WHEN the retrieval process fails
    mock_pdc.retrieve_file_from_pdc.side_effect = subprocess.CalledProcessError(1, "echo")
    with pytest.raises(subprocess.CalledProcessError):
        backup_api.fetch_flow_cell(flow_cell=mock_flow_cell)

    # THEN the failure to retrieve is logged
    assert "retrieval failed" in caplog.text


@mock.patch("cg.meta.backup.backup.BackupAPI.unlink_files")
@mock.patch("cg.meta.backup.backup.BackupAPI.create_rta_complete")
@mock.patch("cg.meta.backup.backup.BackupAPI.query_pdc_for_flow_cell")
@mock.patch("cg.meta.backup.backup.BackupAPI.get_archived_encryption_key_path")
@mock.patch("cg.meta.backup.backup.BackupAPI.get_archived_flow_cell_path")
@mock.patch("cg.meta.tar.tar.TarAPI")
@mock.patch("cg.store.models.Flowcell")
@mock.patch("cg.store")
def test_fetch_flow_cell_integration(
    mock_store,
    mock_flow_cell,
    mock_tar,
    mock_query,
    archived_key,
    archived_flow_cell,
    cg_context,
    pdc_query,
    caplog,
):
    """Component integration test for the BackupAPI, fetching a specified flow cell"""

    caplog.set_level(logging.INFO)

    # GIVEN we want to retrieve a specific flow cell from PDC
    backup_api = BackupAPI(
        encryption_api=mock.Mock(),
        encrypt_dir=cg_context.backup.encrypt_dir.dict(),
        status=mock_store,
        tar_api=mock_tar,
        pdc_api=mock.Mock(),
        root_dir=cg_context.backup.root.dict(),
    )
    mock_flow_cell.status = FlowCellStatus.REQUESTED
    mock_flow_cell.sequencer_type = Sequencers.NOVASEQ
    mock_store.get_flow_cells_by_statuses.return_value.count.return_value = 0
    mock_query.return_value = pdc_query

    backup_api.tar_api.run_tar_command.return_value = None
    result = backup_api.fetch_flow_cell(flow_cell=mock_flow_cell)

    # THEN the process to retrieve the flow cell from PDC is started
    assert "retrieving from PDC" in caplog.text

    # AND when done the status of that flow cell is set to "retrieved"
    assert (
        f"Status for flow cell {mock_flow_cell.name} set to {FlowCellStatus.RETRIEVED}"
        in caplog.text
    )
    assert mock_flow_cell.status == "retrieved"

    # AND status-db is updated with the new status
    assert mock_store.commit.called

    # AND the elapsed time of the retrieval process is returned
    assert result > 0


@mock.patch("cg.meta.backup.backup.SpringBackupAPI.is_spring_file_archived")
@mock.patch("cg.meta.backup.backup.SpringBackupAPI.remove_archived_spring_file")
@mock.patch("cg.meta.backup.backup.SpringBackupAPI.mark_file_as_archived")
@mock.patch("cg.apps.housekeeper.hk.HousekeeperAPI")
@mock.patch("cg.meta.encryption.encryption.SpringEncryptionAPI")
@mock.patch("cg.meta.backup.pdc.PdcAPI")
def test_encrypt_and_archive_spring_file(
    mock_pdc_api: PdcAPI,
    mock_spring_encryption_api: SpringEncryptionAPI,
    mock_housekeeper: HousekeeperAPI,
    mock_mark_file_as_archived,
    mock_remove_archived_spring_files,
    mock_is_archived,
    spring_file_path,
):
    # GIVEN a spring file that needs to be encrypted and archived to PDC and that is not already
    # archived
    spring_backup_api = SpringBackupAPI(
        encryption_api=mock_spring_encryption_api, hk_api=mock_housekeeper, pdc_api=mock_pdc_api
    )
    mock_is_archived.return_value = False

    # WHEN running the encryption and archiving process
    mock_spring_encryption_api.encrypted_spring_file_path.return_value = (
        spring_file_path.with_suffix(FileExtensions.SPRING + FileExtensions.GPG)
    )
    mock_spring_encryption_api.encrypted_key_path.return_value = spring_file_path.with_suffix(
        FileExtensions.KEY + FileExtensions.GPG
    )
    spring_backup_api.encrypt_and_archive_spring_file(spring_file_path)

    # THEN the spring file directory should be cleaned up first
    mock_spring_encryption_api.cleanup.assert_called_with(spring_file_path)

    # AND the spring file should be encrypted
    mock_spring_encryption_api.spring_symmetric_encryption.assert_called_once_with(spring_file_path)

    # AND the key should be encrypted
    mock_spring_encryption_api.key_asymmetric_encryption.assert_called_once_with(spring_file_path)

    # AND the encryption result should be checked
    mock_spring_encryption_api.compare_spring_file_checksums.assert_called_once_with(
        spring_file_path
    )

    # AND the encrypted spring file AND the encrypted key should be archived
    calls = [
        call(
            file_path=str(mock_spring_encryption_api.encrypted_spring_file_path.return_value),
            dry_run=False,
        ),
        call(
            file_path=str(mock_spring_encryption_api.encrypted_key_path.return_value),
            dry_run=False,
        ),
    ]
    mock_pdc_api.archive_file_to_pdc.assert_has_calls(calls)

    # AND the spring file should be marked as archived in Housekeeper
    mock_mark_file_as_archived.assert_called_once_with(spring_file_path)

    # AND the original spring file should be removed
    mock_remove_archived_spring_files.assert_called_once_with(spring_file_path)


@mock.patch("cg.meta.backup.backup.SpringBackupAPI.is_spring_file_archived")
@mock.patch("cg.apps.housekeeper.hk")
@mock.patch("cg.meta.encryption.encryption")
@mock.patch("cg.meta.backup.pdc")
def test_encrypt_and_archive_spring_file_pdc_archiving_failed(
    mock_pdc: PdcAPI,
    mock_spring_encryption_api: SpringEncryptionAPI,
    mock_housekeeper: HousekeeperAPI,
    mock_is_archived,
    spring_file_path,
    caplog,
):
    # GIVEN a spring file that needs to be encrypted and archived to PDC
    spring_backup_api = SpringBackupAPI(
        encryption_api=mock_spring_encryption_api, hk_api=mock_housekeeper, pdc_api=mock_pdc
    )

    # WHEN running the encryption and archiving process, and the encryption command fails
    mock_is_archived.return_value = False
    mock_spring_encryption_api.encrypted_spring_file_path.return_value = (
        spring_file_path.with_suffix(FileExtensions.SPRING + FileExtensions.GPG)
    )
    mock_spring_encryption_api.encrypted_key_path.return_value = spring_file_path.with_suffix(
        FileExtensions.KEY + FileExtensions.GPG
    )
    mock_pdc.archive_file_to_pdc.side_effect = subprocess.CalledProcessError(1, "echo")
    spring_backup_api.encrypt_and_archive_spring_file(spring_file_path=spring_file_path)

    # THEN the appropriate message should be logged and the spring file directory should be
    # cleaned up
    assert "Encryption failed" in caplog.text
    mock_spring_encryption_api.cleanup.assert_called_with(spring_file_path)


@mock.patch("cg.meta.backup.backup.SpringBackupAPI.is_spring_file_archived")
@mock.patch("cg.apps.housekeeper.hk")
@mock.patch("cg.meta.encryption.encryption")
@mock.patch("cg.meta.backup.pdc")
def test_encrypt_and_archive_spring_file_checksum_failed(
    mock_pdc_api: PdcAPI,
    mock_spring_encryption_api: SpringEncryptionAPI,
    mock_housekeeper: HousekeeperAPI,
    mock_is_archived,
    spring_file_path,
    caplog,
):
    # GIVEN a spring file that needs to be encrypted and archived to PDC
    spring_backup_api = SpringBackupAPI(
        encryption_api=mock_spring_encryption_api, hk_api=mock_housekeeper, pdc_api=mock_pdc_api
    )

    # WHEN running the encryption and archiving process
    mock_is_archived.return_value = False
    mock_spring_encryption_api.encrypted_spring_file_path.return_value = (
        spring_file_path.with_suffix(FileExtensions.SPRING + FileExtensions.GPG)
    )
    mock_spring_encryption_api.encrypted_key_path.return_value = spring_file_path.with_suffix(
        FileExtensions.KEY + FileExtensions.GPG
    )

    mock_spring_encryption_api.compare_spring_file_checksums.side_effect = ChecksumFailedError(
        "Checksum comparison failed!"
    )

    spring_backup_api.encrypt_and_archive_spring_file(spring_file_path)

    # THEN the checksum failure should be logged
    assert "Checksum comparison failed!" in caplog.text


@mock.patch("cg.apps.housekeeper.hk")
@mock.patch("cg.meta.encryption.encryption")
@mock.patch("cg.meta.backup.pdc")
def test_mark_file_as_archived(
    mock_pdc_api: PdcAPI,
    mock_spring_encryption_api: SpringEncryptionAPI,
    mock_housekeeper: HousekeeperAPI,
    spring_file_path,
    caplog,
):
    caplog.set_level(logging.INFO)
    # GIVEN a file
    spring_backup_api = SpringBackupAPI(
        encryption_api=mock_spring_encryption_api, hk_api=mock_housekeeper, pdc_api=mock_pdc_api
    )
    mock_housekeeper_file = MockFile(id=0, path=spring_file_path, to_archive=False)
    mock_housekeeper.files.return_value.first.return_value = mock_housekeeper_file

    # WHEN marking the file as archived
    spring_backup_api.mark_file_as_archived(spring_file_path)

    # THEN the set_to_archive method in the Housekeeper API should be called with the new value True
    mock_housekeeper.set_to_archive.assert_called_once_with(file=mock_housekeeper_file, value=True)
    assert f"Setting {spring_file_path} to archived in Housekeeper" in caplog.text


@mock.patch("cg.apps.housekeeper.hk")
@mock.patch("cg.meta.encryption.encryption")
@mock.patch("cg.meta.backup.pdc")
def test_mark_file_as_archived_dry_run(
    mock_pdc_api: PdcAPI,
    mock_spring_encryption_api: SpringEncryptionAPI,
    mock_housekeeper: HousekeeperAPI,
    caplog,
    spring_file_path,
):
    caplog.set_level(logging.INFO)
    # GIVEN a file and running as a dry run
    spring_backup_api = SpringBackupAPI(
        encryption_api=mock_spring_encryption_api,
        hk_api=mock_housekeeper,
        pdc_api=mock_pdc_api,
        dry_run=True,
    )
    mock_housekeeper_file = MockFile(id=0, path=spring_file_path, to_archive=False)
    mock_housekeeper.files.return_value.first.return_value = mock_housekeeper_file

    # WHEN marking the file as archived
    spring_backup_api.mark_file_as_archived(spring_file_path)

    # THEN the set_to_archive method in the Housekeeper API should be called with the new value True
    mock_housekeeper.set_to_archive.assert_not_called()
    assert f"Dry run, no changes made to {spring_file_path}" in caplog.text


@mock.patch("cg.apps.housekeeper.hk")
@mock.patch("cg.meta.encryption.encryption")
@mock.patch("cg.meta.backup.pdc")
def test_decrypt_and_retrieve_spring_file(
    mock_pdc_api: PdcAPI,
    mock_spring_encryption_api: SpringEncryptionAPI,
    mock_housekeeper: HousekeeperAPI,
    spring_file_path,
):
    # GIVEN a spring file that needs to be decrypted and retrieved from PDC
    spring_backup_api = SpringBackupAPI(
        encryption_api=mock_spring_encryption_api, hk_api=mock_housekeeper, pdc_api=mock_pdc_api
    )

    # WHEN running the decryption and retrieval process
    mock_spring_encryption_api.encrypted_spring_file_path.return_value = (
        spring_file_path.with_suffix(FileExtensions.SPRING + FileExtensions.GPG)
    )
    mock_spring_encryption_api.encrypted_key_path.return_value = spring_file_path.with_suffix(
        FileExtensions.KEY + FileExtensions.GPG
    )
    spring_backup_api.retrieve_and_decrypt_spring_file(spring_file_path)

    # THEN the encrypted spring file AND the encrypted key should be retrieved
    calls = [
        call(file_path=str(mock_spring_encryption_api.encrypted_spring_file_path.return_value)),
        call(file_path=str(mock_spring_encryption_api.encrypted_key_path.return_value)),
    ]
    mock_pdc_api.retrieve_file_from_pdc.assert_has_calls(calls)


@mock.patch("cg.apps.housekeeper.hk")
@mock.patch("cg.meta.encryption.encryption")
@mock.patch("cg.meta.backup.pdc")
def test_decrypt_and_retrieve_spring_file_pdc_retrieval_failed(
    mock_pdc: PdcAPI,
    mock_spring_encryption_api: SpringEncryptionAPI,
    mock_housekeeper: HousekeeperAPI,
    spring_file_path,
    caplog,
):
    # GIVEN a spring file that needs to be encrypted and archived to PDC
    spring_backup_api = SpringBackupAPI(
        encryption_api=mock_spring_encryption_api, hk_api=mock_housekeeper, pdc_api=mock_pdc
    )

    # WHEN running the encryption and archiving process
    mock_spring_encryption_api.encrypted_spring_file_path.return_value = (
        spring_file_path.with_suffix(FileExtensions.SPRING + FileExtensions.GPG)
    )
    mock_spring_encryption_api.encrypted_key_path.return_value = spring_file_path.with_suffix(
        FileExtensions.KEY + FileExtensions.GPG
    )
    mock_pdc.retrieve_file_from_pdc.side_effect = subprocess.CalledProcessError(1, "echo")
    spring_backup_api.retrieve_and_decrypt_spring_file(spring_file_path=spring_file_path)

    # THEN the decryption failure should be logged
    assert "Decryption failed" in caplog.text
